import os
import pyscreenshot
from ..training.training import Training
from tkinter import Label, LabelFrame
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from . import decision_trees
from . import bagging_trees
from . import support_vector
from . import neural_network
from . import gaussian_process
from . import nearest_neighbors
from ..styles import colors
from ..utils.indicator import get_indicator_code
from ..utils.insert_canvas import insert_canvas_toolbar 

class View():
    def save_parameter(self):
        #salvar_paramt
        img = pyscreenshot.grab(bbox=(0, 25, 1920, 1040))
        img.show()

        path = os.path.join(os.getcwd(), 'teste.png')  # Caminho atual
        img.save(path)

    def data_preview(self, master, score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis):
        master.laf_res = LabelFrame(master, text='Preview dos resultados', width=1250, height=950, font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=650, y=50)
        Label(master, text='Pontuação (0-100): ' + str(score) + 'pts', font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=70)
        mean_abs_error = round(mean_abs_error, 4)
        Label(master, text='Média Erro absoluto: ' + str(mean_abs_error), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=100)
        mean_rel_error = round(mean_rel_error, 4)
        Label(master, text='Média Erro relativo: ' + str(mean_rel_error), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=130)

        Label(master, text='Maior erro absoluto: ' + str(round(max_abs_error, 4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=160)
        Label(master, text="Valor exato do maior EA: " + str(round(exact_max, 4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=940, y=160)
        Label(master, text="Predict do maior EA: " + str(round(pred_max, 4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=1200, y=160)

        Label(master, text='Menor erro absoluto: ' + str(round(min_abs_error, 4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=680, y=190)
        Label(master, text="Valor exato do menor EA: " + str(round(exact_min, 4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=940, y=190)
        Label(master, text="Predict do menor EA: " + str(round(pred_min, 4)), font='Arial 12 bold', fg='white', bg=colors.fundo).place(x=1200, y=190)

        figure = Figure(figsize=(12, 7.3), dpi=100)
        plot_r = figure.add_subplot(111)
        plot_r.plot(x_axis, y_exact, label='Exato', color='green')
        plot_r.plot(x_axis, y_pred, label='Predict', color='red')
        plot_r.legend()
        plot_r.grid(True)
        plot_r.set_ylabel("Temperature(°C)")
        plot_r.set_xlabel("Comparações")

        canvas = FigureCanvasTkAgg(figure, master=master)
        insert_canvas_toolbar(canvas, master)

    def generate_preview_dt(self, master):
        prev = Training()
        save_model_flag = master.save_model.get()

        # city = master.get_end(master.data_s.get())
        city = master.data_s.get()
        indicator = master.ind_s.get()
        split_percentage = int(master.por_trei.get())
        criterion = master.param_frame.criterion_v.get()
        splitter = master.param_frame.splitter_v.get()
        max_depth = int(master.param_frame.maxd_v.get())
        min_samples_split = master.int_float(master.param_frame.minsam_s_v.get())
        min_samples_leaf = master.int_float(master.param_frame.minsam_l_v.get())
        min_weight_fraction_leaf = float(master.param_frame.minweifra_l_v.get())
        max_features = master.valid_maxf(master.param_frame.maxfeat_v.get())
        max_leaf_nodes = int(master.param_frame.maxleaf_n.get())

        print("Max features ", max_features)
        
        min_impurity_decrease = float(master.param_frame.minimp_dec.get())
        ccp_alpha = float(master.param_frame.ccp_alp_v.get())
        n_tests = int(master.num_teste.get())

        indicator = get_indicator_code(indicator)

        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.decision_tree(
            city, indicator, split_percentage, criterion, splitter, max_depth,
            min_samples_leaf, max_features, max_leaf_nodes, n_tests, min_samples_split,
            min_weight_fraction_leaf, min_impurity_decrease, ccp_alpha, save_model_flag
        )
        master.data_preview(
            score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max,
            min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis
        )

    def generate_preview_bt(self, master):
        prev = Training()
        save_model_flag = master.save_model.get()

        # city = master.get_end(master.data_s.get())
        city = master.data_s.get()
        indicator = master.ind_s.get()
        split_percentage = int(master.por_trei.get())
        criterion = master.param_frame.criterion_v.get()
        splitter = master.param_frame.splitter_v.get()
        max_depth = int(master.param_frame.maxd_v.get())
        min_samples_split = master.int_float(master.param_frame.minsam_s_v.get())
        min_samples_leaf = master.int_float(master.param_frame.minsam_l_v.get())
        min_weight_fraction_leaf = float(master.param_frame.minweifra_l_v.get())
        max_features = master.valid_maxf(master.param_frame.maxfeat_v.get())
        max_leaf_nodes = int(master.param_frame.maxleaf_n.get())

        print("Max features ", max_features)
        
        min_impurity_decrease = float(master.param_frame.minimp_dec.get())
        ccp_alpha = float(master.param_frame.ccp_alp_v.get())
        n_tests = int(master.num_teste.get())
        n_estimators = int(master.n_estimators.get())

        indicator = get_indicator_code(indicator)

        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.bagging_trees(
            city, indicator, split_percentage, criterion, splitter, max_depth,
            min_samples_leaf, max_features, max_leaf_nodes, n_tests, min_samples_split,
            min_weight_fraction_leaf, min_impurity_decrease, ccp_alpha, save_model_flag
        )
        master.data_preview(
            score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max,
            min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis
        )

    def generate_preview_nn(self, master):
        prev = Training()
        save_model_flag = master.save_model.get()

        city = master.get_end(master.data_s.get())

        indicator = master.ind_s.get()
        
        indicator = get_indicator_code(indicator)

        split_percentage = int(master.por_trei.get())

        activation = master.param_frame.activation_v.get()
        solver = master.param_frame.solver_v.get()
        alpha = float(master.param_frame.alpha_v.get())
        batch_size = master.param_frame.batch_size_v.get()
        learning_rate = master.param_frame.learning_rate_v.get()
        learning_rate_init = float(master.param_frame.learning_rate_init_v.get())
        power_t = float(master.param_frame.power_t_v.get())
        max_iter = int(master.param_frame.max_iter_v.get())
        shuffle = master.param_frame.shuffle_v.get()
        tol = float(master.param_frame.tol_v.get())
        verbose = master.param_frame.verbose_v.get()
        warm_start = master.param_frame.warm_start_v.get()
        momentum = float(master.param_frame.momentum_v.get())
        nesterovs_momentum = master.param_frame.nesterovs_momentum_v.get()
        early_stopping = master.param_frame.early_stopping_v.get()
        validation_fraction = float(master.param_frame.validation_fraction_v.get())
        beta_1 = float(master.param_frame.beta_1_v.get())
        beta_2 = float(master.param_frame.beta_2_v.get())
        n_iter_no_change = int(master.param_frame.n_iter_no_change_v.get())
        max_fun = int(master.param_frame.max_fun_v.get())
        n_tests = int(master.num_teste.get())

        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.neural_network(
            city, indicator, split_percentage, n_tests, activation, solver, alpha, batch_size,
            learning_rate, learning_rate_init, power_t, max_iter, shuffle, tol, verbose,
            warm_start, momentum, nesterovs_momentum, early_stopping, validation_fraction,
            beta_1, beta_2, n_iter_no_change, max_fun, save_model_flag
        )

        master.data_preview(
            score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max,
            min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis
        )

    def generate_preview_svm(self, master):
        prev = Training()
        save_model_flag = master.save_model.get()

        city = master.get_end(master.data_s.get())

        indicator = master.ind_s.get()
        indicator = get_indicator_code(indicator)

        split_percentage = int(master.por_trei.get())
        n_tests = int(master.num_teste.get())
        kernel = master.kernel_v.get()
        degree = master.degree_v.get()
        gamma = master.gamma_v.get()
        coef0 = float(master.coef0_v.get())
        tol = float(master.tol_v.get())
        c_param = float(master.c_v.get())
        epsilon = float(master.epsilon_v.get())
        shrinking = master.shrinking_v.get()
        cache_size = float(master.cache_size_v.get())
        verbose = master.verbose_v.get()
        max_iter = int(master.maxiter_v.get())

        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.support_vector_regression(
            city, indicator, split_percentage, n_tests, kernel, degree, gamma, coef0,
            tol, c_param, epsilon, shrinking, cache_size, verbose, max_iter, save_model_flag
        )

        master.data_preview(
            score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max,
            min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis
        )

    def generate_preview_GP(self, master):
        prev = Training()
        save_model_flag = master.save_model.get()

        city = master.get_end(master.data_s.get())
        n_tests = int(master.num_teste.get())
        split_percentage = int(master.por_trei.get())
        alpha_gp = master.param_frame.alpha_gp.get()
        kernel_type = master.kernel_type.get()
        normalize_y_gp = master.param_frame.normalize_y_gp.get()
        lenght_scale = master.param_frame.lenght_scale.get()
        nu = master.param_frame.nu.get()
        sigma_0 = master.param_frame.sigma_0.get()
        alpha_rq = master.param_frame.alpha_rq.get()
        n_restarts_op = master.param_frame.n_restarts_op.get()

        indicator = master.ind_s.get()
        indicator = get_indicator_code(indicator)

        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.gaussian_process_regression(
            city, indicator, split_percentage, n_tests, kernel_type, lenght_scale, 
            nu, sigma_0, alpha_rq, alpha_gp, n_restarts_op, normalize_y_gp, save_model_flag
        )

        master.data_preview(
            score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max,
            min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis
        )


    def generate_preview_kn(self, master):
        prev = Training()
        save_model_flag = master.save_model.get()

        city = master.get_end(master.data_s.get())
        n_tests = int(master.num_teste.get())
        split_percentage = int(master.por_trei.get())
        n_neighbors = master.param_frame.n_neighbors_v.get()
        algorithm = master.param_frame.algorithm_v.get()
        leaf_size = master.param_frame.leaf_size_v.get()
        p_value = master.param_frame.p_v.get()
        n_jobs = master.param_frame.n_jobs_v.get()

        if n_jobs.isdigit() == True:
            n_jobs = int(n_jobs)

        indicator = master.ind_s.get()
        indicator = get_indicator_code(indicator)

        score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max, min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis = prev.KNeighbors(
            city, indicator, split_percentage, n_tests, n_neighbors, algorithm, leaf_size, p_value, n_jobs, save_model_flag
        )

        master.data_preview(
            score, mean_abs_error, mean_rel_error, max_abs_error, exact_max, pred_max,
            min_abs_error, exact_min, pred_min, y_exact, y_pred, x_axis
        )

    def generate_param(self, master):
        opcao = master.ml_selected.get()
        if opcao == 'Decision Trees':
            decision_trees.generate_param(master)
        elif opcao == 'Bagged Trees':
            bagging_trees.generate_param(master)
        elif opcao == 'Neural network':
            neural_network.generate_param(master)
        elif opcao == 'Nearest Neighbors':
            nearest_neighbors.generate_param(master)
        elif opcao == 'Support Vector':
            support_vector.generate_param(master)
        elif opcao == 'Gaussian Process':
            gaussian_process.generate_param(master)


import cdsapi
import pandas as pd 
from geopy.geocoders import Nominatim
import zipfile
import glob
import xarray as xr
import os
import matplotlib.pyplot as plt
import requests
import textwrap
import shutil
import sys 


def get_city_coords(city_name, logger=print):
    """Obtém coordenadas da cidade."""
    geolocator = Nominatim(user_agent="my-cds-app", timeout=10)
    try:
        location = geolocator.geocode(city_name)
        if location:
            return location.latitude, location.longitude
        else:
            logger(f"Aviso: Não foi possível encontrar coordenadas para {city_name}.")
            return None, None
    except Exception as e:
        logger(f"Erro ao buscar as coordenadas: {e}")
        return None, None
    
def get_city_elevation(latitude, longitude, logger=print):
    """Obtém elevação das coordenadas."""
    url = f"https://api.open-elevation.com/api/v1/lookup?locations={latitude},{longitude}"
    try:
        response = requests.get(url)
        data = response.json()
        if data and data['results']:
            elevation = data['results'][0]['elevation']
            return elevation
        else:
            logger("Dados de elevação não encontrados.")
            return "Elevation data not found."
    except Exception as e:
        logger(f"Erro ao buscar elevação: {e}")
        return "Error fetching elevation"

# --- Funções de Processamento com logger ---

def download_era5_data(city_name: str, start_date: str, end_date: str, output_folder: str, logger=print):
    """Baixa os dados do ERA5-Land para a pasta de saída especificada."""
    
    lat, lon = get_city_coords(city_name, logger)
    if lat is None:
        logger("Download cancelado. Houve um erro ao tentar encontrar a cidade.")
        return False # Indica falha
    
    # Garante que a pasta de saída exista
    os.makedirs(output_folder, exist_ok=True)
    
    logger(f"Iniciando download para {city_name} (Lat: {lat:.2f}, Lon: {lon:.2f}) para o período de {start_date} a {end_date}...")

    c = cdsapi.Client()
    dates = pd.date_range(start=start_date, end=end_date, freq='D')

    for (year, month), days_in_month in dates.to_series().groupby([dates.year, dates.month]):
        day_list = [d.strftime('%d') for d in days_in_month]
        output_file_month = os.path.join(output_folder, f'data_{year}_{month:02d}.zip')

        logger(f"Baixando dados para {year}-{month:02d}...")
        try:
            c.retrieve(
                'reanalysis-era5-single-levels',
                {
                    'product_type': 'reanalysis', 'format': 'netcdf',
                    'variable': ['2m_temperature', 'total_precipitation'],
                    'year': str(year), 'month': f'{month:02d}', 'day': day_list,
                    'time': ['00:00', '01:00', '02:00', '03:00', '04:00', '05:00',
                             '06:00', '07:00', '08:00', '09:00', '10:00', '11:00',
                             '12:00', '13:00', '14:00', '15:00', '16:00', '17:00',
                             '18:00', '19:00', '20:00', '21:00', '22:00', '23:00'],
                    'area': [lat, lon, lat, lon],
                },
                output_file_month
            )
        except Exception as e:
            logger(f"ERRO: Falha ao baixar dados para {year}-{month:02d}. Erro: {e}")
            return False # Indica falha
    
    logger(f"Download concluído! Dados salvos em '{output_folder}'.")
    return True # Indica sucesso

def unzip_and_merge_all_nc(zip_folder, output_nc_file, extract_folder, logger=print):
    """Extrai e mescla todos os .nc de uma pasta de zips."""
    os.makedirs(extract_folder, exist_ok=True)
    zip_files = glob.glob(os.path.join(zip_folder, '*.zip'))

    if not zip_files:
        logger(f"Nenhum arquivo .zip encontrado na pasta '{zip_folder}'.")
        return

    datasets_to_merge = []
    logger(f"Encontrados {len(zip_files)} arquivos .zip para processar.")

    try:
        for zip_file in sorted(zip_files):
            logger(f"Processando: {os.path.basename(zip_file)}")
            try:
                with zipfile.ZipFile(zip_file, 'r') as zip_ref:
                    zip_ref.extractall(extract_folder)
            except zipfile.BadZipFile:
                logger(f"Aviso: O arquivo '{os.path.basename(zip_file)}' não é um zip válido. Pulando.")
                continue

            nc_files_in_zip = glob.glob(os.path.join(extract_folder, '*.nc'))
            if nc_files_in_zip:
                ds_month = xr.open_mfdataset(nc_files_in_zip, combine='by_coords', engine='netcdf4')
                ds_month = ds_month.drop_vars(['number', 'expver'], errors='ignore')
                datasets_to_merge.append(ds_month)
                for f in nc_files_in_zip:
                    os.remove(f)
        
        if datasets_to_merge:
            logger("\nUnificando os dados de todos os meses...")
            merged_ds = xr.concat(datasets_to_merge, dim='valid_time')
            logger("\n--- Estrutura do Arquivo Unificado Final ---")
            logger(str(merged_ds)) # Converte para string para logar
            merged_ds.to_netcdf(output_nc_file)
            logger(f"\nDados unificados salvos com sucesso em '{output_nc_file}'.")
            merged_ds.close()
        else:
            logger("Nenhum dado foi encontrado para unificar.")
    except Exception as e:
        logger(f"\nOcorreu um erro ao unificar os arquivos .nc: {e}")
        raise # Re-levanta o erro para a GUI
    finally:
        logger("\nIniciando limpeza final dos .nc...")
        for ds in datasets_to_merge:
            ds.close()
        try:
            if os.path.exists(extract_folder) and not os.listdir(extract_folder):
                 os.rmdir(extract_folder)
                 logger("Pasta de extração temporária removida.")
        except OSError as e:
            logger(f"A pasta de extração '{extract_folder}' não pôde ser removida: {e}")

def process_netcdf_to_daily_csv(nc_file, final_csv, station_name, logger=print):
    """Converte o arquivo .nc unificado em um CSV formatado."""
    try:
        ds = xr.open_dataset(nc_file)

        daily_tmax = (ds['t2m'].resample(valid_time='1D').max() - 273.15).isel(latitude=0, longitude=0)
        daily_tmin = (ds['t2m'].resample(valid_time='1D').min() - 273.15).isel(latitude=0, longitude=0)
        daily_precip = (ds['tp'].resample(valid_time='1D').sum() * 1000).isel(latitude=0, longitude=0)

        df_daily = pd.DataFrame({
            'Data Medicao': daily_tmax.valid_time.values,
            'PRECIPITACAO TOTAL, DIARIO(mm)': daily_precip.values,
            'TEMPERATURA MAXIMA, DIARIA(°C)': daily_tmax.values,
            'TEMPERATURA MINIMA, DIARIA(°C)': daily_tmin.values,
            'latitude': ds['latitude'].values[0],
            'longitude': ds['longitude'].values[0]
        })

        format_daily_csv(df_daily, final_csv, station_name, logger=logger) # Passa o logger
        logger(f"Conversão para CSV concluída com sucesso! Salvo como {final_csv}")

    except FileNotFoundError:
        logger(f"Erro: Arquivo '{nc_file}' não encontrado.")
        raise
    except Exception as e:
        logger(f"Ocorreu um erro durante a conversão: {e}")
        raise

def format_daily_csv(daily_data, output_file, station_name, logger=print):
    """Formata e escreve o DataFrame diário para o arquivo CSV final."""
    lat = daily_data['latitude'].iloc[0]
    lon = daily_data['longitude'].iloc[0]
    elevation = get_city_elevation(lat, lon, logger=logger) # Passa o logger

    header_text = textwrap.dedent(f"""\
    Nome: {station_name}
    Codigo Estacao: ERA5-Land
    Latitude: {lat}
    Longitude: {lon}
    Altitude: {elevation}
    Situacao: N/A
    Data Inicial: {daily_data['Data Medicao'].min().strftime('%Y-%m-%d')}
    Data Final: {daily_data['Data Medicao'].max().strftime('%Y-%m-%d')}
    Periodicidade da Medicao: Diaria

    """)

    daily_data_to_save = daily_data.drop(columns=['latitude', 'longitude'])

    with open(output_file, 'w', encoding='utf-8') as f:
        f.write(header_text)
        ordered_columns = ['Data Medicao', 'PRECIPITACAO TOTAL, DIARIO(mm)', 'TEMPERATURA MAXIMA, DIARIA(°C)', 'TEMPERATURA MINIMA, DIARIA(°C)']
        column_header = ";".join(ordered_columns) + ";\n"
        f.write(column_header)

        for index, row in daily_data_to_save.iterrows():
            date_str = pd.to_datetime(row['Data Medicao']).strftime('%Y-%m-%d')
            precip_str = f"{row['PRECIPITACAO TOTAL, DIARIO(mm)']:.5f}".replace('.', ',')
            tmax_str = f"{row['TEMPERATURA MAXIMA, DIARIA(°C)']:.4f}".replace('.', ',')
            tmin_str = f"{row['TEMPERATURA MINIMA, DIARIA(°C)']:.4f}".replace('.', ',')
            line = f"{date_str};{precip_str};{tmax_str};{tmin_str};\n"
            f.write(line)


def download_and_process_era_data(city, start_date, end_date, output_dir, logger=print):
    """
    Função principal que orquestra todo o processo de download e conversão.
    Todos os arquivos são salvos DENTRO do output_dir.
    """
    
    city_slug = city.split(',')[0].replace(' ', '_').lower()
    temp_download_folder = os.path.join(output_dir, f"temp_zip_{city_slug}")
    temp_extract_folder = os.path.join(output_dir, "temp_extract")
    unified_nc_file = os.path.join(output_dir, f"{city_slug}_unificado.nc")
    final_csv_file = os.path.join(output_dir, f"ECMWF_{city_slug}_{start_date}_to_{end_date}.csv") # Nome de arquivo final claro

    os.makedirs(temp_download_folder, exist_ok=True)
    os.makedirs(temp_extract_folder, exist_ok=True)

    logger("="*30)
    logger(f"Iniciando processo para a cidade: {city}")
    logger("="*30)
    
    success = download_era5_data(
        city_name=city,
        start_date=start_date, 
        end_date=end_date,
        output_folder=temp_download_folder,
        logger=logger
    )
    
    if not success:
        logger("Falha na etapa de download. Abortando.")
        return 

    unzip_and_merge_all_nc(
        temp_download_folder, 
        unified_nc_file, 
        temp_extract_folder, 
        logger=logger
    )
    
    process_netcdf_to_daily_csv(
        unified_nc_file, 
        final_csv_file, 
        city,
        logger=logger
    )

    logger(f"\n--- Limpando arquivos temporários para {city} ---")
    try:
        shutil.rmtree(temp_download_folder)
        shutil.rmtree(temp_extract_folder)
        os.remove(unified_nc_file)
        logger("Limpeza concluída com sucesso!")
    except OSError as e:
        logger(f"Erro durante a limpeza: {e}")
    logger("\nProcesso finalizado.")


import pandas as pd
import numpy as np
from scipy import stats

def mann_kendall_test(data):
    n = len(data)
    s = 0
    for i in range(n - 1):
        for j in range(i + 1, n):
            s += np.sign(data[j] - data[i])

    unique, counts = np.unique(data, return_counts=True)
    
    if len(unique) == n:
        tie_correction = 0
    else:
        tie_correction = np.sum(counts[counts > 1] * (counts[counts > 1] - 1) * (2 * counts[counts > 1] + 5))

    variance = (n * (n - 1) * (2 * n + 5) - tie_correction) / 18
    
    if variance > 0:
        if s > 0:
            z = (s - 1) / np.sqrt(variance)
        elif s < 0:
            z = (s + 1) / np.sqrt(variance)
        else: # s == 0
            z = 0
    else:
        return {"trend": "no trend", "p-value": 1.0, "Tau": 0, "Z-score": 0}

    trend = "increasing" if s > 0 else "decreasing" if s < 0 else "no trend"

    p_value = 2 * (1 - stats.norm.cdf(abs(z)))

    tau = s / (n * (n - 1) / 2)
    
    return {"trend": trend, "p-value": p_value, "Tau": tau, "Z-score": z}

def analyze_trend(csv_file, column_name):
    data_base = pd.read_csv(csv_file)
    
    if column_name not in data_base.columns:
        raise ValueError(f"Coluna '{column_name}' não encontrada no CSV.")
    
    data_series = data_base[column_name].dropna().to_numpy()

    mk_result = mann_kendall_test(data_series)

    n = len(data_series)
    slopes = [(data_series[j] - data_series[i]) / (j - i) for i in range(n - 1) for j in range(i + 1, n)]
    sens_slope = np.median(slopes)

    return {
        "mann_kendall": mk_result,
        "sens_slope": sens_slope
    }

def analyze_indicator_trend(file_path):
    column_name = 'annual'
    try:
        data_base = pd.read_csv(
            file_path,
            header=0,
            sep=','
        )
    except Exception as e:
        print(f"ERRO ao ler {file_path}. Ignorando. Erro: {e}")
        return None
    
    if data_base.columns[0] != 'year' and not data_base.columns[0].startswith('Unnamed'):
         data_base.rename(columns={data_base.columns[0]: 'year'}, inplace=True)
    
    if column_name not in data_base.columns:
        print(f"AVISO: Coluna '{column_name}' não encontrada em {file_path}. Colunas disponíveis: {data_base.columns.tolist()}")
        return None
    
    data_series = data_base[column_name].replace('', np.nan).dropna().astype(float).to_numpy()

    if len(data_series) < 3:
        print(f"AVISO: {file_path} tem menos de 3 pontos de dados na coluna '{column_name}' e foi ignorado.")
        return None
    
    mk_result = mann_kendall_test(data_series)

    n = len(data_series)

    slopes = [(data_series[j] - data_series[i]) / (j - i) 
              for i in range(n - 1) 
              for j in range(i + 1, n)]
              
    sens_slope = np.median(slopes)

    return {
        "mann_kendall": mk_result,
        "sens_slope": sens_slope
    }


from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.tree import DecisionTreeRegressor

class CustomDecisionTree(BaseEstimator, RegressorMixin):
    def __init__(self, criterion='squared_error', splitter='best', max_depth=None,
                 min_samples_leaf=1, max_features=None, max_leaf_nodes=None,
                 min_samples_split=2, min_weight_fraction_leaf=0.0,
                 min_impurity_decrease=0.0, ccp_alpha=0.0):
        self.criterion = criterion
        self.splitter = splitter
        self.max_depth = max_depth
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_samples_split = min_samples_split
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha

    def fit(self, X, y):
        self.model_ = DecisionTreeRegressor(
            criterion=self.criterion,
            splitter=self.splitter,
            max_depth=self.max_depth,
            min_samples_leaf=self.min_samples_leaf,
            max_features=self.max_features,
            max_leaf_nodes=self.max_leaf_nodes,
            min_samples_split=self.min_samples_split,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            min_impurity_decrease=self.min_impurity_decrease,
            ccp_alpha=self.ccp_alpha
        )
        self.model_.fit(X, y)
        return self

    def predict(self, X):
        return self.model_.predict(X)



import time
from typing import Tuple, List
import numpy as np
import os
import pickle
from sklearn.neural_network import MLPRegressor
from .utils import calculate_errors

def neural_network(
    self, city: str, indicator_code: int, split_ratio: float, n_tests: int,
    activation: str, solver: str, alpha: float, batch_size, learning_rate: str,
    learning_rate_init: float, power_t: float, max_iter: int, shuffle: bool,
    tol: float, verbose: bool, warm_start: bool, momentum: float,
    nesterovs_momentum: bool, early_stopping: bool, validation_fraction: float,
    beta_1: float, beta_2: float, n_iter_no_change: int, max_fun: int, save_model: bool
) -> Tuple[float, float, float, float, float, float, float, float, float, List[float], List[float], List[int]]:
    """
    Trains and evaluates an MLPRegressor multiple times and returns error statistics.

    Returns:
        Tuple containing:
            score, average absolute error, average relative error,
            max absolute error, exact value at max error, predicted value at max error,
            min absolute error, exact value at min error, predicted value at min error,
            list of exact values, list of predicted values, x axis indices
    """
    start_time = time.time()

    indicators = {3: "Precipitation", 4: 'Maximum temperature'}
    indicator = indicators.get(indicator_code, 'Minimum temperature')

    train_X, train_y, val_X, val_y = self.prepare_matrix_by_city(city, split_ratio, indicator_code)

    all_exact = []
    all_predicted = []
    total_absolute_error = 0
    total_relative_error = 0
    x_axis = []
    counter = 1

    for test in range(n_tests):
        model = MLPRegressor(
            activation=activation, solver=solver, alpha=alpha, batch_size=batch_size,
            learning_rate=learning_rate, learning_rate_init=learning_rate_init, power_t=power_t,
            max_iter=max_iter, shuffle=shuffle, tol=tol, verbose=verbose, warm_start=warm_start,
            momentum=momentum, nesterovs_momentum=nesterovs_momentum, early_stopping=early_stopping,
            validation_fraction=validation_fraction, beta_1=beta_1, beta_2=beta_2,
            n_iter_no_change=n_iter_no_change, max_fun=max_fun
        )

        model.fit(train_X, train_y)

        predictions = model.predict(val_X)
        absolute_errors = np.abs(np.array(val_y) - predictions)
        relative_errors = absolute_errors / np.array(val_y)

        total_absolute_error += np.mean(absolute_errors)
        total_relative_error += np.mean(relative_errors)

        if test != n_tests:  # sempre True — mesma situação da sua condição anterior
            all_exact.extend(val_y)
            all_predicted.extend(predictions)
            x_axis.extend(range(counter, counter + len(val_y)))
            counter += len(val_y)

    statistics = calculate_errors(all_exact, all_predicted, total_absolute_error, total_relative_error, n_tests)

    score = statistics["score"]
    max_error = statistics["max_error"]
    min_error = statistics["min_error"]
    exact_max = statistics["exact_max"]
    predicted_max = statistics["predicted_max"]
    exact_min = statistics["exact_min"]
    predicted_min = statistics["predicted_min"]
    avg_absolute_error = statistics["avg_absolute_error"]
    avg_relative_error = statistics["avg_relative_error"]

    if save_model:
        with open(os.path.join(os.getcwd(), 'modelo_rn.sav'), 'wb') as f:
                pickle.dump(model, f)

    return (
        score, avg_absolute_error, avg_relative_error,
        max_error, exact_max, predicted_max,
        min_error, exact_min, predicted_min,
        all_exact, all_predicted, x_axis
    )



import tkinter as tk
from tkinter import ttk, filedialog, messagebox as msg
import os
import typing
import sys
import subprocess
from datetime import datetime

from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk

from ..utils import *
if typing.TYPE_CHECKING:
    from ..start_page import StartPage
from ...trends.plot_warming_stripes import plot_annual_data, plot_monthly_data, plot_quarterly_data
from ...trends.processing import process_csv
from ...trends.group_data import group_data
from ...trends.identify_trends import analyze_trend
from ...trends.climdex import Climdex
from ..tooltip import CreateToolTip

class ClimateTrendsPage(ttk.Frame):
    """Tela para análise de tendências climáticas."""
    def __init__(self, parent, controller):
        self.city_name = None
        ttk.Frame.__init__(self, parent)
        self.controller = controller

        self.processed_file_name = None
        self.output_dir = None
        self.city_name_raw = None

        top_frame = ttk.Frame(self)
        top_frame.pack(fill=tk.X, padx=10, pady=10)

        self.back_button = ttk.Button(top_frame, text="", command=self.go_to_start_page)
        self.back_button.pack(side=tk.LEFT)

        self.page_title = ttk.Label(top_frame, text="", font=("Verdana", 16, "bold"))
        self.page_title.pack(side=tk.LEFT, expand=True)

        # --- Layout Principal de Duas Colunas ---
        main_container = ttk.Frame(self)
        main_container.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        main_container.grid_columnconfigure(0, weight=1, minsize=280)
        main_container.grid_columnconfigure(1, weight=3)
        main_container.grid_rowconfigure(0, weight=1)
        
        # --- PAINEL DA ESQUERDA ---
        left_panel = ttk.Frame(main_container)
        left_panel.grid(row=0, column=0, sticky="nsew", padx=(0, 10))
        
        # Seção de Processamento de Arquivo
        self.processed_file_label = ttk.Label(left_panel, text="")
        self.processed_file_label.pack(anchor="w", padx=5)

        self.processed_file_entry = ttk.Entry(left_panel, state="readonly")
        self.processed_file_entry.pack(fill=tk.X, padx=5, pady=(0, 5))
        self.processed_file_hint = CreateToolTip(self.processed_file_entry, text='')

        self.select_process_btn = ttk.Button(left_panel, text="", command=self.choose_and_process_file)
        self.select_process_btn.pack(fill=tk.X, padx=5, pady=5)
        self.select_process_hint = CreateToolTip(self.select_process_btn, text='')

        self.extremes_frame = ttk.LabelFrame(left_panel, text="")
        self.extremes_frame.pack(fill=tk.X, padx=5, pady=(10, 5))
        
        # Frame interno para os combos
        date_frame = ttk.Frame(self.extremes_frame)
        date_frame.pack(fill=tk.X, padx=5, pady=5)
        date_frame.grid_columnconfigure(0, weight=1)
        date_frame.grid_columnconfigure(1, weight=1)

        self.start_year_label = ttk.Label(date_frame, text="")
        self.start_year_label.grid(row=0, column=0, sticky="w")
        self.start_year_combo = ttk.Combobox(date_frame, state="disabled")
        self.start_year_combo.grid(row=1, column=0, sticky="ew", padx=(0, 5))
        self.start_hint = CreateToolTip(self.start_year_combo, text='')

        self.end_year_label = ttk.Label(date_frame, text="")
        self.end_year_label.grid(row=0, column=1, sticky="w")
        self.end_year_combo = ttk.Combobox(date_frame, state="disabled")
        self.end_year_combo.grid(row=1, column=1, sticky="ew", padx=(5, 0))
        self.end_hint = CreateToolTip(self.end_year_combo, text='')

        self.climatic_extremes_btn = ttk.Button(left_panel, text="", command=self.calculate_extreme_indicators)
        self.climatic_extremes_btn.pack(fill=tk.X, padx=5, pady=5)
        self.climatic_extremes_hint = CreateToolTip(self.climatic_extremes_btn, text='')

        self.analyze_trend_frame = ttk.LabelFrame(left_panel, text="")
        self.analyze_trend_frame.pack(fill=tk.X, padx=5, pady=(10, 5))

        self.trend_param_label = ttk.Label(self.analyze_trend_frame, text="")
        self.trend_param_label.pack(anchor="w", padx=5)
        
        self.trend_param_combo = ttk.Combobox(self.analyze_trend_frame, state="readonly", values=["Maximum temperature", "Minimum temperature", "Precipitation", "Mean temperature"])
        self.trend_param_combo.pack(fill=tk.X, padx=5, pady=5)
        self.analyze_trend_param_hint = CreateToolTip(self.trend_param_combo, text=''
                                                )
        self.analyze_trend_btn = ttk.Button(self.analyze_trend_frame, text="", command=self.run_trend_analysis)
        self.analyze_trend_btn.pack(fill=tk.X, padx=5, pady=(5, 10))
        self.analyze_trend_hint = CreateToolTip(self.analyze_trend_btn, text='')

        # Seção de Plotagem de Trends
        self.plot_trends_frame = ttk.LabelFrame(left_panel, text="")
        self.plot_trends_frame.pack(fill=tk.X, padx=5)

        self.plot_param_label = ttk.Label(self.plot_trends_frame, text="")
        self.plot_param_label.pack(anchor="w", padx=5)
        
        self.plot_param_combo = ttk.Combobox(self.plot_trends_frame, state="readonly", values= ["Maximum temperature", "Minimum temperature", "Precipitation", "Mean temperature"])
        self.plot_param_combo.pack(fill=tk.X, padx=5, pady=5)
        self.plot_param_hint = CreateToolTip(self.plot_param_combo, text='')

        self.monthly_btn = ttk.Button(self.plot_trends_frame, text="", command=self.plot_monthly)
        self.monthly_btn.pack(fill=tk.X, expand=True, pady=5, padx=5)
        self.quarterly_btn = ttk.Button(self.plot_trends_frame, text="", command=self.plot_quarterly)
        self.quarterly_btn.pack(fill=tk.X, expand=True, pady=5, padx=5)
        self.annual_btn = ttk.Button(self.plot_trends_frame, text="", command=self.plot_annual)
        self.annual_btn.pack(fill=tk.X, expand=True, pady=5, padx=5)
        self.monthly_hint = CreateToolTip(self.monthly_btn, text='')
        self.quarterly_hint = CreateToolTip(self.quarterly_btn, text='')
        self.annual_hint = CreateToolTip(self.annual_btn, text='')

        # --- PAINEL DA DIREITA ---
        self.right_panel = ttk.Frame(main_container, relief="solid", borderwidth=1)
        self.right_panel.grid(row=0, column=1, sticky="nsew")
        
        self.update_texts()

    def _clear_panel(self, panel):
        """Limpa todos os widgets de um painel."""
        for widget in panel.winfo_children():
            widget.destroy()

    def _get_years_from_file(self, file_path):
        """Lê o arquivo processado e retorna uma lista de anos únicos."""
        available_years = set()
        try:
            with open(file_path, 'r') as f:
                for line in f:
                    if not line.strip(): continue
                    parts = line.split(',')
                    if parts[0].isdigit():
                        available_years.add(int(parts[0]))
        except FileNotFoundError:
            self.controller.show_translated_message(
                'error', 'folder_not_found_title', 'folder_not_found_msg', directory=file_path
            )
            return []
        except Exception as e:
            self.controller.show_translated_message(
                'error', 'years_load_error_title', 'years_load_error_msg', error=str(e)
            )
            return []
        
        if not available_years:
            return []
            
        sorted_years = sorted(list(available_years))
        return [str(year) for year in sorted_years]

    def choose_and_process_file(self):
        print("choose and process file")
        """Abre o diálogo, processa o arquivo e atualiza a UI."""
        caminho_completo = filedialog.askopenfilename(
            parent=self,
            title="Selecione o arquivo CSV de dados brutos",
            filetypes=[("Arquivos CSV", "*.csv"), ("Todos os arquivos", "*.*")]
        )

        if not caminho_completo: return

        self.city_name_raw = os.path.basename(caminho_completo)
        
        city_name_only = self.city_name_raw.split('.')[0]
        self.city_name = city_name_only
        
        self.city_name_raw = os.path.basename(caminho_completo)
        input_dir = os.path.dirname(caminho_completo)
        base_dir = os.path.abspath(os.path.dirname(sys.argv[0]))
        self.output_dir = os.path.join(base_dir, "output_dir")
        os.makedirs(self.output_dir, exist_ok=True)
        self.city = [self.city_name_raw]

        try:
            process_csv(self.city, input_dir, self.output_dir)
            self.processed_file_name = f"{self.city_name_raw}"

            self.processed_file_entry.config(state="normal")
            self.processed_file_entry.delete(0, tk.END)
            self.processed_file_entry.insert(0, self.processed_file_name)
            self.processed_file_entry.config(state="readonly")
            
            processed_path = os.path.join(self.output_dir, self.processed_file_name)
            years_list = self._get_years_from_file(processed_path)
            
            if years_list:
                self.start_year_combo.config(values=years_list, state="readonly")
                self.end_year_combo.config(values=years_list, state="readonly")
                # Define valores padrão (primeiro e último ano)
                self.start_year_combo.set(years_list[0])
                self.end_year_combo.set(years_list[-1])
            else:
                self.start_year_combo.config(values=[], state="disabled")
                self.end_year_combo.config(values=[], state="disabled")

            self.controller.show_translated_message('info', 'success_title', 'file_processed_success_msg')
        
        except Exception as e:
            self.controller.show_translated_message('error', 'processing_error_title', 'processing_error_msg', error=str(e))
    
    def calculate_extreme_indicators(self):
        """Calcula e mostra os indicadores climáticos."""
        print("calculate extreme indicators")
        i18n = self.controller.i18n
        if not self.processed_file_name:
            self.controller.show_translated_message(
                'warning', 'file_not_processed_title', 'file_not_processed_msg'
            )
            return
        
        start_year_str = self.start_year_combo.get()
        end_year_str = self.end_year_combo.get()

        if not start_year_str or not end_year_str:
            self.controller.show_translated_message('warning', 'year_selection_error_title', 'year_selection_error_msg')
            return

        start_year = int(start_year_str)
        end_year = int(end_year_str)

        if end_year < start_year:
            self.controller.show_translated_message('warning', 'year_selection_error_title', 'year_order_error_msg')
            return

        start_date = datetime(start_year, 1, 1)
        end_date = datetime(end_year, 12, 31)
        
        try:
            c = Climdex()
            data = c.read_files_climdex(self.output_dir, self.city)
            city_name_with_ext = self.city[0]
            city_name_only = city_name_with_ext.split('.')[0]
            self.city_name = city_name_only
            df_city = data[city_name_with_ext]

            indices = c.calculate_indices(df_city, (start_date, end_date)) # Usa as datas dos combos
            c.write_indices(indices, self.city_name, self.output_dir)

            pdf_output_path = f"{self.output_dir}/graphs_indices_{self.city_name}.pdf"
            c.plot_and_save_indices(indices, self.city_name, pdf_output_path)

            self._display_file_list(self.output_dir)

        except Exception as e:
            self.controller.show_translated_message('error', 'calc_indices_error_title', 'calc_indices_error_msg', error=str(e))

    def _plot_graph_on_panel(self, fig):
        print("plot graph on panel")
        """Limpa o painel direito e desenha uma nova figura matplotlib nele."""
        for widget in self.right_panel.winfo_children():
            widget.destroy()

        if fig is None:
            return

        canvas = FigureCanvasTkAgg(fig, master=self.right_panel)
        canvas.draw()

        toolbar = NavigationToolbar2Tk(canvas, self.right_panel)
        toolbar.update()
        
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)
        toolbar.pack(side=tk.BOTTOM, fill=tk.X)

    def _display_file_list(self, directory):
        """Limpa o painel direito e exibe uma lista de arquivos de um diretório."""
        i18n = self.controller.i18n
        
        self._clear_panel(self.right_panel)

        try:
            # Pega a lista de arquivos
            files = os.listdir(directory)
            # Filtra para mostrar apenas arquivos (sem pastas) e talvez tipos específicos
            files = [f for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f)) and not f.startswith('.')]
        except FileNotFoundError:
            self.controller.show_translated_message(
                'error', 'folder_not_found_title', 'folder_not_found_msg', directory=directory
            )
            return
        except Exception as e:
            self.controller.show_translated_message(
                'error', 'file_list_error_title', 'file_list_error_msg', error=str(e)
            )
            return

        list_frame = ttk.Frame(self.right_panel)
        list_frame.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        
        label = ttk.Label(list_frame, text=i18n.get('generated_files_label'), font=("Verdana", 10, "bold"))
        label.pack(anchor="w", pady=(0, 5))

        scrollbar = ttk.Scrollbar(list_frame, orient=tk.VERTICAL)
        
        self.file_tree = ttk.Treeview(
            list_frame,
            columns=("filename"),
            show="headings",
            yscrollcommand=scrollbar.set
        )
        scrollbar.config(command=self.file_tree.yview)

        self.file_tree.heading("filename", text=i18n.get('filename_header'))
        self.file_tree.column("filename", anchor="w")

        for f in sorted(files):
            self.file_tree.insert("", "end", values=(f,))

        # Empacota os widgets no frame
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.file_tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        # Evento de clique duplo para abrir o arquivo
        def open_file_location(event):
            try:
                item_id = self.file_tree.selection()[0]
                filename = self.file_tree.item(item_id, "values")[0]
                full_path = os.path.join(directory, filename)
                
                # Abre o arquivo com o programa padrão do sistema
                if sys.platform == "win32":
                    os.startfile(full_path)
                elif sys.platform == "darwin": # macOS
                    subprocess.run(["open", full_path])
                else: # Linux
                    subprocess.run(["xdg-open", full_path])
            except IndexError:
                pass # Clique duplo fora de um item
            except Exception as e:
                 self.controller.show_translated_message(
                    'error', 'open_file_error_title', 'open_file_error_msg', file=filename, error=str(e)
                )

        self.file_tree.bind("<Double-1>", open_file_location)

    def get_parameter(self, temperature_parameter):
        print("get parameter")
        if temperature_parameter == "Maximum temperature":
            parameter = "tmax"
        elif temperature_parameter == "Minimum temperature":
            parameter = "tmin"
        elif temperature_parameter == "Precipitation":
            parameter = "prec"
        else:
            parameter = "tmean"
        return parameter
    
    def plot_monthly(self):
        print("Entrou entrou plot_monthly")
        city_name = self.city[0].split('.')
        city_name = city_name[0]
        self.city_name = city_name

        group_data(self.city, self.output_dir, self.output_dir)
        
        temperature_parameter = self.plot_param_combo.get()
        if not temperature_parameter:
            self.controller.show_translated_message('warning', 'param_not_selected_title', 'param_not_selected_msg')
            return
        
        parameter_abreviation = self.get_parameter(temperature_parameter)
        
        fig = plot_monthly_data(
            csv_path=f'{self.output_dir}/{self.city_name}DadosMensais.csv',
            index=parameter_abreviation,  # Coluna a ser plotada ('tmax', 'tmin', 'tmean', 'prec')
            file_name=f'tendencia_mensal_{parameter_abreviation}_{self.city_name}.png',
            title_img=f'Tendência da {temperature_parameter} Mensal em {self.city_name}',
            caption_img='',
            embed_mode=True
        )

        self._plot_graph_on_panel(fig)
    
    def plot_quarterly(self):
        city_name = self.city[0].split('.')
        city_name = city_name[0]
        self.city_name = city_name

        group_data(self.city, self.output_dir, self.output_dir)
        
        temperature_parameter = self.plot_param_combo.get()
        if not temperature_parameter:
            self.controller.show_translated_message('warning', 'param_not_selected_title', 'param_not_selected_msg')
            return
        
        parameter_abreviation = self.get_parameter(temperature_parameter)
        
        fig = plot_quarterly_data(
            csv_path=f'{self.output_dir}/{self.city_name}dadosTrimestrais.csv',
            index=parameter_abreviation,  # Coluna a ser plotada ('tmax', 'tmin', 'tmean', 'prec')
            file_name=f'tendencia_trimestral_{parameter_abreviation}_{self.city_name}.png',
            title_img=f'Tendência da {temperature_parameter} Trimestral em {self.city_name}',
            caption_img='',
            embed_mode=True
        )

        self._plot_graph_on_panel(fig)

    def plot_annual(self):
        city_name = self.city[0].split('.')
        city_name = city_name[0]
        self.city_name = city_name

        group_data(self.city, self.output_dir, self.output_dir)

        temperature_parameter = self.plot_param_combo.get()
        if not temperature_parameter:
            self.controller.show_translated_message('warning', 'param_not_selected_title', 'param_not_selected_msg')
            return
        
        parameter_abreviation = self.get_parameter(temperature_parameter)
        
        fig = plot_annual_data(
            csv_path=f'{self.output_dir}/{self.city_name}dadosAnuais.csv',
            index=parameter_abreviation,  # Coluna a ser plotada ('tmax', 'tmin', 'tmean', 'prec')
            file_name=f'tendencia_anual_{parameter_abreviation}_{self.city_name}.png',
            title_img=f'Tendência da {temperature_parameter} Anual em {self.city_name}',
            caption_img='',
            embed_mode=True
        )
        self._plot_graph_on_panel(fig)

    def run_trend_analysis(self):
        """Coleta dados, chama a análise de tendência e exibe os resultados."""
        i18n = self.controller.i18n
        
        if not self.processed_file_name or not self.city_name:
            self.controller.show_translated_message('warning', 'file_not_processed_title', 'file_not_processed_msg')
            return
        
        temperature_parameter = self.trend_param_combo.get()
        if not temperature_parameter:
            self.controller.show_translated_message('warning', 'param_not_selected_title', 'param_not_selected_msg')
            return
        
        try:
            group_data(self.city, self.output_dir, self.output_dir)
            
            csv_file_path = f'{self.output_dir}/{self.city_name}dadosAnuais.csv' # Foco na tendência anual
            column_name = self.get_parameter(temperature_parameter) # ex: 'tmax', 'tmin', 'tmean'
            
            if column_name == "tmax":
                column_name = "Tmax"
            elif column_name == "tmin":
                column_name = "Tmin"
            elif column_name == "tmean":
                column_name = "Tmean"
            elif column_name == "Precipitation" or column_name == "prec":
                column_name = "Chuva"

            if not os.path.exists(csv_file_path):
                 self.controller.show_translated_message('error', 'file_not_found_title', 'annual_data_not_found_msg', file_path=csv_file_path)
                 return

            results = analyze_trend(csv_file_path, column_name)
            
            self._display_trend_results(results, temperature_parameter)

        except ValueError as e:
            self.controller.show_translated_message('error', 'trend_analysis_error_title', 'trend_analysis_value_error_msg', error=str(e))
        except Exception as e:
            self.controller.show_translated_message('error', 'trend_analysis_error_title', 'trend_analysis_error_msg', error=str(e))

    def _display_trend_results(self, results, param_name):
        """Limpa o painel direito e exibe os resultados da análise de tendência."""
        self._clear_panel(self.right_panel)
        i18n = self.controller.i18n

        results_frame = ttk.LabelFrame(self.right_panel, text=i18n.get('trend_results_title'))
        results_frame.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        
        grid_frame = ttk.Frame(results_frame)
        grid_frame.pack(padx=10, pady=10)

        mk_result = results.get("mann_kendall", {})
        print("MK:", mk_result)
        slope = results.get("sens_slope")

        trend = mk_result.get('trend', 'no trend')
        p_value = mk_result.get('p-value', 1.0)
        tau = mk_result.get('Tau', 0.0)
        z_score = mk_result.get('Z-score', 0.0)
        
        if trend == 'increasing':
            trend_text = i18n.get('trend_increasing')
        elif trend == 'decreasing':
            trend_text = i18n.get('trend_decreasing')
        else: # no trend
            trend_text = i18n.get('trend_no_trend')

        row_num = 0

        # Parâmetro
        ttk.Label(grid_frame, text=f"{i18n.get('parameter_label')}:", font=("Verdana", 10, "bold")).grid(row=row_num, column=0, sticky="w", pady=2); row_num += 1
        ttk.Label(grid_frame, text=param_name).grid(row=0, column=1, sticky="w", pady=2, padx=5)
        
        # Teste Mann-Kendall (Tendência)
        ttk.Label(grid_frame, text=f"{i18n.get('mann_kendall_test_label')}:", font=("Verdana", 10, "bold")).grid(row=row_num, column=0, sticky="w", pady=2); row_num += 1
        ttk.Label(grid_frame, text=trend_text).grid(row=row_num-1, column=1, sticky="w", pady=2, padx=5)

        # p-value
        ttk.Label(grid_frame, text=f"{i18n.get('p_value_label')}:", font=("Verdana", 10, "bold")).grid(row=row_num, column=0, sticky="w", pady=2); row_num += 1
        ttk.Label(grid_frame, text=f"{p_value:.6f}").grid(row=row_num-1, column=1, sticky="w", pady=2, padx=5)

        ttk.Label(grid_frame, text=f"{i18n.get('tau')}:", font=("Verdana", 10, "bold")).grid(row=row_num, column=0, sticky="w", pady=2); row_num += 1
        ttk.Label(grid_frame, text=f"{tau:.6f}").grid(row=row_num-1, column=1, sticky="w", pady=2, padx=5)

        ttk.Label(grid_frame, text=f"{i18n.get('Z-score')}:", font=("Verdana", 10, "bold")).grid(row=row_num, column=0, sticky="w", pady=2); row_num += 1
        ttk.Label(grid_frame, text=f"{z_score:.6f}").grid(row=row_num-1, column=1, sticky="w", pady=2, padx=5)
        
        ttk.Label(grid_frame, text=f"{i18n.get('sens_slope_label')}:", font=("Verdana", 10, "bold")).grid(row=row_num, column=0, sticky="w", pady=2); row_num += 1
        ttk.Label(grid_frame, text=f"{slope:.6f} (unidades/ano)").grid(row=row_num-1, column=1, sticky="w", pady=2, padx=5)
        
        ttk.Separator(grid_frame, orient='horizontal').grid(row=row_num, column=0, columnspan=2, sticky='ew', pady=10); row_num += 1
        explanation = i18n.get('trend_explanation')
        ttk.Label(grid_frame, text=explanation, wraplength=450, justify=tk.LEFT).grid(row=row_num, column=0, columnspan=2, sticky="w")

    def go_to_start_page(self):
        from ..start_page import StartPage
        self.controller.show_frame(StartPage)
        
    def update_texts(self):
        i18n = self.controller.i18n
        self.back_button.config(text=i18n.get('back_btn'))
        self.page_title.config(text=i18n.get('climate_trends_page_title'))
        self.processed_file_label.config(text=i18n.get('processed_file_label'))
        self.select_process_btn.config(text=i18n.get('select_and_process_btn'))
        self.extremes_frame.config(text=i18n.get('climate_extremes_label'))
        self.start_year_label.config(text=i18n.get('start_year_label'))
        self.end_year_label.config(text=i18n.get('end_year_label'))
        self.climatic_extremes_btn.config(text=i18n.get('climatic_extremes_btn'))
        self.analyze_trend_frame.config(text=i18n.get('analyze_trend_label'))
        self.trend_param_label.config(text=i18n.get('trend_parameter_label'))
        self.analyze_trend_btn.config(text=i18n.get('analyse_trend_btn'))
        
        self.plot_trends_frame.config(text=i18n.get('plot_trends_label'))
        self.plot_param_label.config(text=i18n.get('plot_parameter_label'))
        self.analyze_trend_btn.config(text=i18n.get('analyse_trend_btn'))
        self.plot_trends_frame.config(text=i18n.get('plot_trends_label'))
        self.monthly_btn.config(text=i18n.get('monthly_btn'))
        self.quarterly_btn.config(text=i18n.get('quarterly_btn'))
        self.annual_btn.config(text=i18n.get('annual_btn'))

        self.select_process_hint.text = i18n.get('select_process_trends_hint')
        self.processed_file_hint.text = i18n.get('processed_file_trends_hint')
        self.start_hint.text = i18n.get('start_trends_hint')
        self.end_hint.text = i18n.get('end_trends_hint')
        self.climatic_extremes_hint.text = i18n.get('climatic_trends_hint')
        self.analyze_trend_param_hint.text = i18n.get('analyze_param_trends_hint')
        self.analyze_trend_hint.text = i18n.get('analyze_trends_hint')
        self.plot_param_hint.text = i18n.get('param_trends_hint')
        self.monthly_hint.text = i18n.get('monthly_trends_hint')
        self.quarterly_hint.text = i18n.get('quarterly_trends_hint')
        self.annual_hint.text = i18n.get('monthly_trends_hint') 



from .utils import *
from .gui_helpers import NNParameterFrame

def generate_param(self):
    self.param_frame = NNParameterFrame(self, fundo)
    self.data_s = StringVar()
    self.data_s.set('Target city')
    lista_dt = ['Target city', 'Neighbor A', 'Neighbor B', 'Neighbor C']
    Label(self, text="Dados para treinamento:", font='Arial 12 bold', fg='white', bg=fundo).place(x=50, y=750)
    self.combo_c = ttk.Combobox(self, values=lista_dt, textvariable=self.data_s, width=25, font='Arial 12', justify=CENTER, state='readonly').place(x=50, y=775)

    self.ind_s = StringVar()
    self.ind_s.set('Maximum temperature')
    lista_ind = ["Precipitation", 'Maximum temperature', 'Minimum temperature']
    Label(self, text='Indicador:', font='Arial 12 bold', fg='white', bg=fundo).place(x=340, y=750)
    ttk.Combobox(self, values=lista_ind, textvariable=self.ind_s, width=25, font='Arial 12', justify=CENTER, state='readonly').place(x=340, y=775)

    self.por_trei = IntVar()
    self.por_trei.set(70)
    Label(self, text="Porção para treinamento:", font='Arial 12 bold', fg='white', bg=fundo).place(x=50, y=810)
    Scale(self, variable=self.por_trei, orient=HORIZONTAL, length=240).place(x=50, y=835)

    self.num_teste = IntVar()
    self.num_teste.set(5)
    Label(self, text="Número de testes (int):", font='Arial 12 bold', fg='white', bg=fundo).place(x=340, y=810)
    self.ent_num_teste = Entry(self, textvariable=self.num_teste, width=27, font='Arial 12', justify=CENTER).place(x=340, y=835)

    Button(self, text='Preview', font='Arial 11 bold', fg='white', bg=fun_b, width=25, command=self.generate_preview_nn).place(x=50, y=915)
    self.save_model = IntVar()
    Checkbutton(self, text='Salvar modelo', variable=self.save_model, bg=fundo, font='Arial 12 bold', activebackground=fundo).place(x=340, y=915)


    


from sklearn import tree
from sklearn.neural_network import MLPRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR
from math import floor

def customized_meta_learning(self, indicator, base_l, triangulation_method, meta_l, pre1, pre2, n_test, sliding_window):
        m1_40_t, m1_40_r, m2_40_t, m2_40_r, m3_20_t, m3_20_r = self.prepare_input(indicator, sliding_window)

        if base_l != 'None':
            input_matrix, base_ea, base_er, base_percentage, base_r2 = self.base_learn(
                base_l, 0, n_test, m1_40_t, m1_40_r, m3_20_t, m3_20_r, m2_40_t, m2_40_r, sliding_window
            )
            del input_matrix[:2]

        if triangulation_method != 'None':
            _, triangulation_matrix, tria_ea, tria_er = self.triangula(triangulation_method, indicator)
            del triangulation_matrix[:4]

        data_matrix, _, _, _ = self.triangula('Arithmetic Average', indicator)
        del data_matrix[:4]
        del m2_40_r[:2]

        final_matrix = []
        final_results = []

        for i in range(len(data_matrix)):
            aux = []
            if base_l == 'None':
                aux.extend([data_matrix[i][0], data_matrix[i][1], data_matrix[i][2], triangulation_matrix[i]])
            elif triangulation_method == 'None':
                aux.extend([data_matrix[i][0], data_matrix[i][1], data_matrix[i][2], input_matrix[i][3]])
            else:
                aux.extend([data_matrix[i][0], data_matrix[i][1], data_matrix[i][2], triangulation_matrix[i], input_matrix[i][3]])
            final_matrix.append(aux)

        if pre2 == 0:
            if meta_l == 'Decision Trees':
                level1_learner = tree.DecisionTreeRegressor()
            elif meta_l == 'Neural Network':
                level1_learner = MLPRegressor()
            elif meta_l == 'Nearest Neighbors':
                level1_learner = KNeighborsRegressor()
            elif meta_l == 'Support Vector':
                level1_learner = SVR()

        total_size = len(final_matrix)
        train_limit = total_size - floor(total_size * 0.2)

        x_train = []
        y_train = []
        x_test = []
        y_test = []

        for i in range(total_size):
            if i <= train_limit:
                x_train.append(final_matrix[i])
                y_train.append(m2_40_r[i])
            else:
                x_test.append(final_matrix[i])
                y_test.append(m2_40_r[i])

        r2_sum = 0
        total_absolute_error = 0
        total_relative_error = 0

        x_meta = []
        y_meta = []
        y_target = []
        x_counter = 0

        for _ in range(n_test):
            level1_learner = level1_learner.fit(x_train, y_train)
            r2_sum += level1_learner.score(x_test, y_test)

            fold_absolute_error = 0
            fold_relative_error = 0

            for j in range(len(x_test)):
                expected_value = float(y_test[j])
                predicted_value = float(level1_learner.predict([x_test[j]])[0])

                y_meta.append(predicted_value)
                y_target.append(expected_value)
                x_meta.append(x_counter)
                x_counter += 1

                absolute_error = abs(expected_value - predicted_value)
                relative_error = absolute_error / expected_value

                fold_absolute_error += absolute_error
                fold_relative_error += relative_error

            total_absolute_error += fold_absolute_error / len(x_test)
            total_relative_error += fold_relative_error / len(x_test)

        meta_ea = total_absolute_error / n_test
        meta_er = total_relative_error / n_test
        meta_percentage_error = meta_ea * 100
        meta_r2 = r2_sum / n_test

        if base_l == 'None':
            return meta_ea, meta_er, meta_percentage_error, meta_r2, x_meta, y_meta, y_target, 0, 0, 0, 0, tria_ea, tria_er
        elif triangulation_method == 'None':
            return meta_ea, meta_er, meta_percentage_error, meta_r2, x_meta, y_meta, y_target, base_ea, base_er, base_percentage, base_r2, 0, 0
        else:
            return meta_ea, meta_er, meta_percentage_error, meta_r2, x_meta, y_meta, y_target, base_ea, base_er, base_percentage, base_r2, tria_ea, tria_er



from math import floor
import pyscreenshot
import os
from sklearn import tree
from sklearn.neural_network import MLPRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR
from ..data_processing.data_processing import DataProcessing
from ..training.training import Training
from ..triangulation.triangulation import Triangulation
from .view import View
from .gen_param import generate_param
from .customized_meta_l import customized_meta_learning
from .combine_meta_l import combine_meta_learning
from ..styles import colors


class MetaLearning:
    def prepare_input(self, indicator, sliding_window):
        if indicator == 1:
            focus_column = 3
        elif indicator == 2:
            focus_column = 4
        else: 
            focus_column = 5

        data_processor = DataProcessing()
        raw_data = data_processor.load_data_file('Common data')

        data_matrix = []
        for row in raw_data:  # Build a matrix with year, month, day and focus value
            processed_row = [
                float(row[0]),
                float(row[1]),
                float(row[2]),
                float(row[focus_column])
            ]
            data_matrix.append(processed_row)

        normalizer = Training()
        normalized_data = normalizer.normalize(data_matrix)  # Normalize each column and each row

        if sliding_window == 'Yes':
            input_matrix, target_matrix = self.sliding_window(normalized_data)  # Prepare data using sliding window format
        else:
            input_matrix, target_matrix = self.common_input(normalized_data)

        model1_inputs, model1_targets = [], []
        model2_inputs, model2_targets = [], []  # For level-0 learning to make predictions, which will be used by the meta-learner
        model3_inputs, model3_targets = [], []

        total_samples = len(input_matrix)
        split_point1 = floor(total_samples * 0.4)
        split_point2 = split_point1 * 2

        for index in range(total_samples):  # Split the data into portions
            if index <= split_point1:
                model1_inputs.append(input_matrix[index])
                model1_targets.append(target_matrix[index])
            elif split_point1 < index <= split_point2:
                model2_inputs.append(input_matrix[index])
                model2_targets.append(target_matrix[index])
            else:
                model3_inputs.append(input_matrix[index])
                model3_targets.append(target_matrix[index])

        return (
            model1_inputs, model1_targets,
            model2_inputs, model2_targets,
            model3_inputs, model3_targets
        )
    
    def sliding_window(self, data):
        matrix = []
        result = []

        for i in range(len(data)):
            buff = []
            try:
                for j in range(5):
                    buff.append(data[i+j][0])
                    buff.append(data[i+j][1])
                    buff.append(data[i+j][2])
                    buff.append(data[i+j][3])
                if len(buff) == 20:
                    matrix.append(buff[:19])
                    result.append(buff[19])
            except IndexError:
                pass
        return matrix, result
    
    def common_input(self, data):
        matrix = list()
        result = list()
        
        for i in range(len(data)):
            buff = list()
            buff.append(data[i][0])
            buff.append(data[i][1])
            buff.append(data[i][2])
            matrix.append(buff)

            result.append(data[i][3])

        return matrix, result
    
    def base_learn(self, model_name, pretrained, num_tests,
               train_inputs, train_targets,
               validation_inputs, validation_targets,
               predict_inputs, predict_targets, sliding_window):

        if pretrained == 0:
            if model_name == 'Decision Trees':
                base_learner = tree.DecisionTreeRegressor()
            elif model_name == 'Neural network':
                base_learner = MLPRegressor()
            elif model_name == 'Nearest Neighbors':
                base_learner = KNeighborsRegressor()
            elif model_name == 'Support Vector':
                base_learner = SVR()

        total_relative_error = 0
        total_absolute_error = 0
        total_r2_score = 0

        for test_index in range(num_tests):
            base_learner = base_learner.fit(train_inputs, train_targets)
            
            fold_absolute_error = 0
            fold_relative_error = 0
            total_r2_score += base_learner.score(validation_inputs, validation_targets)

            for sample_index in range(len(validation_inputs)):
                expected_value = float(validation_targets[sample_index])
                predicted_value = float(base_learner.predict([validation_inputs[sample_index]])[0])
                
                absolute_error = abs(expected_value - predicted_value)
                relative_error = absolute_error / expected_value

                fold_absolute_error += absolute_error
                fold_relative_error += relative_error

            total_absolute_error += fold_absolute_error / len(validation_inputs)
            total_relative_error += fold_relative_error / len(validation_inputs)

        mean_absolute_error = total_absolute_error / num_tests
        mean_relative_error = total_relative_error / num_tests
        error_percentage = mean_absolute_error * 100
        mean_r2_score = total_r2_score / num_tests

        # Preparing level-0 learner predictions for level-1 learner
        prepared_predict_matrix = []
        
        if sliding_window == 'Yes':
            for i in range(len(predict_inputs)):
                predicted_value = float(base_learner.predict([predict_inputs[i]])[0])
                prepared_sample = [
                    predict_inputs[i][16],
                    predict_inputs[i][17],
                    predict_inputs[i][18],
                    predicted_value
                ]
                prepared_predict_matrix.append(prepared_sample)
        else:
            for i in range(len(predict_inputs)):
                predicted_value = float(base_learner.predict([predict_inputs[i]])[0])
                prepared_sample = [
                    predict_inputs[i][0],
                    predict_inputs[i][1],
                    predict_inputs[i][2],
                    predicted_value
                ]
                prepared_predict_matrix.append(prepared_sample)

        return prepared_predict_matrix, mean_absolute_error, mean_relative_error, error_percentage, mean_r2_score
    
    def triangula(self, metodo, focus):
        triang = Triangulation()
        nor = Training()
        if metodo == 'Inverse Distance Weighted':
            triang.idw(focus)
            matriz_triang = nor.normalize(triang.get_idw()[5])
            x,y,alv_y, erro_abs, erro_rel, mat_ext = triang.get_idw()
            erro_abs, erro_rel = self.calculate_error_tri(alv_y, y)
        elif metodo == 'Arithmetic Average':
            triang.avg(focus)
            matriz_triang = nor.normalize(triang.get_avg()[5])
            x,y,alv_y, erro_abs, erro_rel, mat_ext = triang.get_avg()
            erro_abs, erro_rel = self.calculate_error_tri(alv_y, y)
        elif metodo == 'Regional Weight':
            triang.rw(focus)
            matriz_triang = nor.normalize(triang.get_rw()[5])
            x,y,alv_y, erro_abs, erro_rel, mat_ext = triang.get_rw()
            erro_abs, erro_rel = self.calculate_error_tri(alv_y, y)
        elif metodo == 'Optimized Normal Ratio':
            triang.onr(focus)
            matriz_triang = nor.normalize(triang.get_onr()[5])
            x,y,alv_y, erro_abs, erro_rel, mat_ext = triang.get_onr()
            erro_abs, erro_rel = self.calculate_error_tri(alv_y, y)

        tamanho = len(matriz_triang)
        t1 = floor(tamanho * 0.4)
        t2 = t1 * 2

        matriz_final_data = list()
        matriz_final_dado = list()
        
        for i in range(len(matriz_triang)):
            if i > t1 and i <= t2:
                aux = list()
                aux.append(matriz_triang[i][0])
                aux.append(matriz_triang[i][1])
                aux.append(matriz_triang[i][2])
                matriz_final_data.append(aux)
                matriz_final_dado.append(matriz_triang[i][3])
                

        return matriz_final_data, matriz_final_dado, erro_abs, erro_rel
       
    def calculate_error_tri(self, x, y):
        training = Training()
        mat1 = training.normalize(x)
        mat2 = training.normalize(y)
        
        sum_ea = 0
        sum_er = 0
        for i in range(len(mat1)):
            ea = abs(float(mat1[i]) - float(mat2[i]))
            er = ea / float(mat1[i])

            sum_ea += ea
            sum_er += er

        ea = sum_ea / len(mat1)
        er = sum_er / len(mat2)
        return ea, er
    
    def customized_meta_learning(self, indicator, base_l, triangulation_method, meta_l, pre1, pre2, n_test, sliding_window):
        return customized_meta_learning(self, indicator, base_l, triangulation_method, meta_l, pre1, pre2, n_test, sliding_window)

    def combine_meta_learning(self, target, pre1, pre2, n_test, window):
        return combine_meta_learning(self, target, pre1, pre2, n_test, window)

    def valid_maxf(self, val):
        if val.isdigit() == True:
            val = int(val)
        elif val.isalnum() == True and val.isdigit() == False:
            val = str(val)
        elif val.isalnum() == False and val.isdigit() == False and val.isalpha() == False:
            val = float(val)
        
        return val
    
    def salvar_paramt(self):
        img = pyscreenshot.grab(bbox=(0, 25, 1920, 1040))
        img.show()
        
        path = os.path.join(os.getcwd(), 'teste.png')
        img.save(path)

    def data_preview(self, pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x):
        v = View()
        v.data_preview(self, pts, media_ea, media_er, maior_ea, exat_maior, pre_maior, menor_ea, exat_menor, pre_menor, eixo_y_exato, eixo_y_predict, eixo_x)

    def get_end(self, cidade):
        treatment = DataProcessing()
        return treatment.get_file_path(cidade)

    def generate_preview_dt(self):
        v = View()
        v.generate_preview_dt(self)

    def generate_preview_bt(self):
        v = View()
        v.generate_preview_bt(self)
    
    def generate_preview_nn(self):
        """
        Generates a preview of the model Nearest Neighbors
        """
        v = View()
        v.generate_preview_nn(self)
    
    def generate_preview_svm(self):
        """
        Generates a preview of the model Support Vector Machine
        """
        v = View()
        v.generate_preview_svm(self)
    
    def generate_preview_Kn(self):
        """
        Generates a preview o K Neighbors
        """
        v = View()
        v.generate_preview_Kn

    def generate_param(self):
        generate_param(self)


from .utils import *
from .gui_helpers import BaggingTParameterFrame

def generate_param(self):
    Canvas(self, width=615, height=900, background=fundo, border=0).place(x=10, y=95)
    self.param_frame = BaggingTParameterFrame(self, fundo, fun_alt)
    create_training_widgets(self)
    Button(self, text='Preview', font='Arial 11 bold', fg='white', bg=fun_b, width=25, command=self.generate_preview_bt).place(x=50, y=685)
    self.save_model = IntVar()
    Checkbutton(self, text='Save model', variable=self.save_model, bg=fundo, font='Arial 12 bold', activebackground=fundo).place(x=340, y=685)


import pandas as pd
import xarray as xr
import xclim
from xclim.indicators import icclim
from xclim import indices
from pathlib import Path
import os
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np


class Climdex:
    def __init__(self):
        self.read_columns = ["year", "month", "day", "pr", "tmax", "tmin"]
        self.write_columns = ["year", "jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec", "annual"]
        
        self.indices_base = [
            "TXx", "TX10p", "TX90p", "TNn", "TN10p", "TN90p",
            "PRCPTOT", "R95p", "RX1DAY", "RX5DAY", "CDD", "CWD"
        ]
        
        self.units_map = {
            "TXx": "°C", "TX10p": "dias", "TX90p": "dias",
            "TNn": "°C", "TN10p": "dias", "TN90p": "dias",
            "PRCPTOT": "mm", "R95p": "mm", "RX1DAY": "mm", "RX5DAY": "mm",
            "CDD": "dias", "CWD": "dias"
        }

    def read_files_climdex(self, processed_dir: str, station_names: list) -> dict:
        dataframes = {}
        col_names = ["year", "month", "day", "prec", "tmax", "tmin", "tmean"]

        for name in station_names:
            file_path = os.path.join(processed_dir, f"{name}")
            
            if not os.path.exists(file_path):
                print(f"AVISO: Arquivo não encontrado em '{file_path}'. Pulando estação '{name}'.")
                continue

            df = pd.read_csv(file_path, header=None, names=col_names)

            rename_dict = {
                'prec': 'pr',
                'tmax': 'tasmax',
                'tmin': 'tasmin'
            }
            df.rename(columns=rename_dict, inplace=True)
            
            df['time'] = pd.to_datetime(df[['year', 'month', 'day']])
            df.set_index('time', inplace=True)
            df.drop(columns=['year', 'month', 'day'], inplace=True)

            if 'Unnamed: 0' in df.columns:
                df.drop(columns=['Unnamed: 0'], inplace=True)

            dataframes[name] = df
            
        return dataframes

    def calculate_indices(self, df: pd.DataFrame, base_period: tuple[str, str]):
        """
        Calcula uma lista de índices climáticos do ETCCDI usando xclim.
        """
        # Converter o DataFrame do pandas para um Dataset do xarray
        if not isinstance(df.index, pd.DatetimeIndex):
            raise ValueError("O DataFrame de entrada deve ter um DatetimeIndex.")
        ds = df.to_xarray()

        ds.tasmax.attrs.update({
            'standard_name': 'air_temperature',
            'cell_methods': 'time: maximum',
            'units': 'degC'
        })
        ds.tasmin.attrs.update({
            'standard_name': 'air_temperature',
            'cell_methods': 'time: minimum',
            'units': 'degC'
        })
        if 'pr' in ds:
            ds.pr.attrs.update({
                'standard_name': 'precipitation_flux',
                'cell_methods': 'time: mean',
                'units': 'mm/day'
            })

        # Preparar o período base para cálculo dos percentis
        ds_base = ds.sel(time=slice(base_period[0], base_period[1]))

        # Pré-calcular todos os percentis necessários
        with xr.set_options(keep_attrs=True):
            tx10_thresh = xclim.core.calendar.percentile_doy(ds_base.tasmax, per=10).sel(percentiles=10).drop_vars('percentiles')
            tx90_thresh = xclim.core.calendar.percentile_doy(ds_base.tasmax, per=90).sel(percentiles=90).drop_vars('percentiles')
            tn10_thresh = xclim.core.calendar.percentile_doy(ds_base.tasmin, per=10).sel(percentiles=10).drop_vars('percentiles')
            tn90_thresh = xclim.core.calendar.percentile_doy(ds_base.tasmin, per=90).sel(percentiles=90).drop_vars('percentiles')
            if 'pr' in ds:
                pr95_thresh = xclim.core.calendar.percentile_doy(ds_base.pr, per=95).sel(percentiles=95).drop_vars('percentiles')

        #  Definir e calcular cada índice com os nomes de função corretos
        indices_results = {}
        freq_m = 'MS'  # Frequência Mensal (Month Start)

        # --- Índices de Temperatura ---
        indices_results['TXx'] = icclim.TXx(tasmax=ds.tasmax, freq=freq_m)
        indices_results['TNn'] = icclim.TNn(tasmin=ds.tasmin, freq=freq_m)
        indices_results['TX10p'] = indices.tx10p(tasmax=ds.tasmax, tasmax_per=tx10_thresh, freq=freq_m)
        indices_results['TX90p'] = indices.tx90p(tasmax=ds.tasmax, tasmax_per=tx90_thresh, freq=freq_m)
        indices_results['TN10p'] = indices.tn10p(tasmin=ds.tasmin, tasmin_per=tn10_thresh, freq=freq_m)
        indices_results['TN90p'] = indices.tn90p(tasmin=ds.tasmin, tasmin_per=tn90_thresh, freq=freq_m)
        indices_results['TR'] = icclim.TR(tasmin=ds.tasmin, freq=freq_m)
        indices_results['FD'] = icclim.FD(tasmin=ds.tasmin, freq=freq_m)


        # --- Índices de Precipitação ---
        if 'pr' in ds:
            indices_results['PRCPTOT'] = icclim.PRCPTOT(pr=ds.pr, freq=freq_m)
            indices_results['R95p'] = icclim.R95p(pr=ds.pr, pr_per=pr95_thresh, freq=freq_m)
            indices_results['RX1DAY'] = icclim.RX1day(pr=ds.pr, freq=freq_m)
            indices_results['RX5DAY'] = icclim.RX5day(pr=ds.pr, freq=freq_m)
            indices_results['CDD'] = icclim.CDD(pr=ds.pr, freq=freq_m)
            indices_results['CWD'] = icclim.CWD(pr=ds.pr, freq=freq_m)
            indices_results['R10mm'] = icclim.R10mm(pr=ds.pr, freq=freq_m)
            indices_results['R20mm'] = icclim.R20mm(pr=ds.pr, freq=freq_m)

        else:
            print("Aviso: Coluna 'pr' de precipitação não encontrada. Índices de precipitação não serão calculados.")

        # Combinar todos os resultados em um único DataFrame
        valid_indices = {k: v for k, v in indices_results.items() if v is not None}
        final_ds = xr.Dataset(valid_indices)
        
        freq_a = 'AS'  # Frequência Anual (Year Start)
        annual_indices = {}

        # Agregação para índices de contagem ou soma (dias, mm)
        sum_indices = ["TX10p", "TX90p", "TN10p", "TN90p", "PRCPTOT", "R95p", "R10mm", "R20mm", "FD", "TR", "CWD", "CDD"]
        for name, da in final_ds.items():
            if name in sum_indices:
                annual_indices[f"{name}_annual"] = da.resample(time=freq_a).sum(skipna=False)

        # Agregação para índices de máximo (°C, mm)
        max_indices = ["TXx", "RX1DAY", "RX5DAY"]
        for name, da in final_ds.items():
            if name in max_indices:
                annual_indices[f"{name}_annual"] = da.resample(time=freq_a).max(skipna=False)

        # Agregação para índices de mínimo (°C)
        min_indices = ["TNn"]
        for name, da in final_ds.items():
            if name in min_indices:
                annual_indices[f"{name}_annual"] = da.resample(time=freq_a).min(skipna=False)
                
        # 7. Combinar os resultados anuais com os mensais
        final_ds = xr.merge([final_ds, xr.Dataset(annual_indices)])

        return final_ds

    def write_indices(self, indices_ds: xr.Dataset, name: str, output_dir: str = "indices_xclim"):
        output_path = Path(output_dir)
        output_path.mkdir(parents=True, exist_ok=True)
        temp_indices_to_correct = ["TXx", "TNn"]
        for index_name in self.indices_base:
            monthly_da = indices_ds[index_name]
            annual_da = indices_ds[f"{index_name}_annual"]

            if index_name in temp_indices_to_correct:
                monthly_da = monthly_da - 273.15
                annual_da = annual_da - 273.15
            df_monthly = monthly_da.to_dataframe()
            df_wide = df_monthly.pivot_table(values=index_name, index=df_monthly.index.year, columns=df_monthly.index.month).rename_axis('year', axis='index').rename_axis(None, axis='columns')
            month_names = [pd.to_datetime(f"2024-{i}-01").strftime('%b').lower() for i in range(1, 13)]
            df_wide.columns = month_names
            df_annual = annual_da.to_dataframe().rename(columns={f"{index_name}_annual": "annual"})
            df_annual.index = df_annual.index.year
            final_df = df_wide.join(df_annual['annual']).reset_index().reindex(columns=self.write_columns)
            final_csv_path = output_path / f"{name}_{index_name}"
            final_df.round(2).to_csv(final_csv_path, index=False, sep=",")
            print(f"Índice salvo em: {final_csv_path}")

    def plot_and_save_indices(self, indices_ds: xr.Dataset, name: str, output_dir: str = "graficos_indices"):
        output_path = Path(output_dir)
        output_path.mkdir(parents=True, exist_ok=True)

        temp_indices_to_correct = ["TXx", "TNn"]
        
        for index_name in self.indices_base:
            print(f"Gerando gráfico para o índice: {index_name}")
            
            monthly_da = indices_ds[index_name].copy(deep=True)
            if index_name in temp_indices_to_correct:
                monthly_da = monthly_da - 273.15

            df_long = monthly_da.to_dataframe(name=index_name).reset_index()
            df_long['year'] = df_long['time'].dt.year
            df_long['month'] = df_long['time'].dt.month

            pdf_path = output_path / f"{name}_decadal_{index_name}.pdf"

            with PdfPages(pdf_path) as pdf:
                start_year = df_long["year"].min()
                end_year = df_long["year"].max()
                
                decades = list(range(start_year, end_year + 1, 10))
                if not decades: continue

                plt.figure(figsize=(12, len(decades) * 2.5))

                for i, start_decade in enumerate(decades):
                    end_decade = start_decade + 9
                    subset = df_long[(df_long["year"] >= start_decade) & (df_long["year"] <= end_decade)]

                    if subset.empty:
                        continue

                    ax = plt.subplot(len(decades), 1, i + 1)
                    
                    x_axis = subset["year"] + (subset["month"] - 1) / 12
                    ax.plot(x_axis, subset[index_name], color="blue", linewidth=0.8, marker='o', markersize=2, linestyle='-')

                    unit = self.units_map.get(index_name, "")
                    ax.set_title(f"Estação: {name}, Década: {start_decade}-{min(end_decade, end_year)}, Índice: {index_name}", fontsize=10)
                    ax.set_ylabel(f"Valor ({unit})")
                    
                    ax.set_xlim(start_decade, min(end_decade, end_year) + 1)
                    ax.set_xticks(range(start_decade, min(end_decade, end_year) + 2, 1))
                    
                    ax.grid(True, linestyle='--', alpha=0.6)

                plt.tight_layout()
                pdf.savefig()
                plt.close()
            
            print(f"Gráfico salvo em: {pdf_path}")


import tkinter as tk
from tkinter import ttk, messagebox as msg, StringVar, IntVar, BooleanVar
import typing
import os
import threading

from ...tooltip import CreateToolTip
from ....meta_learning.meta_learning import MetaLearning

from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from tksheet import Sheet

if typing.TYPE_CHECKING:
    from ...start_page import StartPage

class MetaLearningPage(ttk.Frame):
    """Tela para configuração e execução de Meta Learning."""
    def __init__(self, parent, controller):
        ttk.Frame.__init__(self, parent)
        self.controller = controller

        self.ml_lv0_var = StringVar(value='Decision Trees')
        self.ml_tr0_var = StringVar(value='Arithmetic Average')
        self.ml_lv1_var = StringVar(value='Decision Trees')
        self.ind_meta_perso_var = StringVar(value='Maximum temperature')
        self.num_teste_mtp_var = IntVar(value=1)
        self.pre_para_lv0_var = IntVar(value=0)
        self.pre_para_lv1_var = IntVar(value=0)
        self.type_input_var = StringVar(value='Yes')

        self.pre_nn_comb_var = IntVar(value=0)
        self.pre_dt_comb_var = IntVar(value=0)
        self.pre_bt_comb_var = IntVar(value=0)
        self.pre_nneig_comb_var = IntVar(value=0)
        self.pre_sv_comb_var = IntVar(value=0)
        self.pre_gp_comb_var = IntVar(value=0)
        self.ind_meta_comb_var = StringVar(value='Maximum temperature')
        self.num_teste_mtc_var = IntVar(value=1)

        # --- Flags de Animação ---
        self.loading_ct = False
        self.loading_gt = False

        # --- Frame Superior ---
        top_frame = ttk.Frame(self)
        top_frame.pack(fill=tk.X, padx=10, pady=10)
        self.back_button = ttk.Button(top_frame, text="", command=self.go_to_previous_page)
        self.back_button.pack(side=tk.LEFT)
        self.page_title = ttk.Label(top_frame, text="", font=("Verdana", 16, "bold"))
        self.page_title.pack(side=tk.LEFT, expand=True)

        # --- Layout Principal ---
        main_container = ttk.Frame(self)
        main_container.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        main_container.grid_columnconfigure(0, weight=1, minsize=400) # Painel esquerdo maior
        main_container.grid_columnconfigure(1, weight=3) # Painel direito
        main_container.grid_rowconfigure(0, weight=1)

        # --- PAINEL ESQUERDO (Com Scrollbar) ---
        left_panel_base = ttk.Frame(main_container)
        left_panel_base.grid(row=0, column=0, sticky="nsew", padx=(0, 10))
        left_panel_base.grid_rowconfigure(0, weight=1)
        left_panel_base.grid_columnconfigure(0, weight=1)

        canvas_left = tk.Canvas(left_panel_base)
        canvas_left.grid(row=0, column=0, sticky="nsew")
        scrollbar_left = ttk.Scrollbar(left_panel_base, orient="vertical", command=canvas_left.yview)
        scrollbar_left.grid(row=0, column=1, sticky="ns")
        canvas_left.configure(yscrollcommand=scrollbar_left.set)
        self.scrollable_frame_left = ttk.Frame(canvas_left)
        canvas_left.create_window((0, 0), window=self.scrollable_frame_left, anchor="nw")
        self.scrollable_frame_left.bind("<Configure>", lambda e: canvas_left.configure(scrollregion=canvas_left.bbox("all")))
        canvas_left.bind_all("<MouseWheel>", lambda e: canvas_left.yview_scroll(int(-1*(e.delta/120)), "units")) # Windows
        canvas_left.bind_all("<Button-4>", lambda e: canvas_left.yview_scroll(-1, "units")) # Linux up
        canvas_left.bind_all("<Button-5>", lambda e: canvas_left.yview_scroll(1, "units")) # Linux down

        custom_test_frame = ttk.LabelFrame(self.scrollable_frame_left, text="")
        custom_test_frame.pack(fill=tk.X, padx=5, pady=5)
        self.custom_test_frame_label = custom_test_frame

        self.lbl_base_learning = ttk.Label(custom_test_frame, text="")
        self.lbl_base_learning.pack(anchor="w", padx=5)
        lista_ml0 = ['None','Decision Trees', 'Bagged Trees', 'Neural network', 'Nearest Neighbors', 'Support Vector', 'Gaussian Process']
        self.ml0 = ttk.Combobox(custom_test_frame, values=lista_ml0, textvariable=self.ml_lv0_var, state='readonly')
        self.ml0.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.ml0_hint = CreateToolTip(self.ml0, text='')

        self.lbl_triangulation = ttk.Label(custom_test_frame, text="")
        self.lbl_triangulation.pack(anchor="w", padx=5)
        lista_tr0 = ['None', 'Arithmetic Average', 'Inverse Distance Weighted', 'Regional Weight', 'Optimized Normal Ratio']
        self.tr0 = ttk.Combobox(custom_test_frame, values=lista_tr0, textvariable=self.ml_tr0_var, state='readonly')
        self.tr0.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.tr0_hint = CreateToolTip(self.tr0, text='')

        self.lbl_meta_learning = ttk.Label(custom_test_frame, text="")
        self.lbl_meta_learning.pack(anchor="w", padx=5)
        lista_ml1 = ['Decision Trees', 'Bagged Trees', 'Neural network', 'Nearest Neighbors', 'Support Vector', 'Gaussian Process']
        self.ml1 = ttk.Combobox(custom_test_frame, values=lista_ml1, textvariable=self.ml_lv1_var, state='readonly')
        self.ml1.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.ml1_hint = CreateToolTip(self.ml1, text='')

        self.lbl_indicator_ct = ttk.Label(custom_test_frame, text="")
        self.lbl_indicator_ct.pack(anchor="w", padx=5)
        lista_ind_meta_p = ["Precipitation", 'Maximum temperature', 'Minimum temperature']
        self.indicator_ct = ttk.Combobox(custom_test_frame, values=lista_ind_meta_p, textvariable=self.ind_meta_perso_var, state='readonly')
        self.indicator_ct.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.indicator_ct_hint = CreateToolTip(self.indicator_ct, text='')

        self.lbl_num_tests_ct = ttk.Label(custom_test_frame, text="")
        self.lbl_num_tests_ct.pack(anchor="w", padx=5)
        self.num_tests = ttk.Entry(custom_test_frame, textvariable=self.num_teste_mtp_var, justify=tk.CENTER)
        self.num_tests.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.num_tests_hint = CreateToolTip(self.num_tests, text='')

        self.lbl_pretrained = ttk.Label(custom_test_frame, text="")
        self.lbl_pretrained.pack(anchor="w", padx=5)
        self.pretrained_hint = CreateToolTip(self.lbl_pretrained, text='')
        pretrained_frame = ttk.Frame(custom_test_frame)
        pretrained_frame.pack(fill=tk.X, padx=5)
        self.cb_pre_lv0 = ttk.Checkbutton(pretrained_frame, text="", variable=self.pre_para_lv0_var, onvalue=1, offvalue=0)
        self.cb_pre_lv0.pack(side=tk.LEFT, padx=(0, 10))
        self.cb_pre_lv1 = ttk.Checkbutton(pretrained_frame, text="", variable=self.pre_para_lv1_var, onvalue=1, offvalue=0)
        self.cb_pre_lv1.pack(side=tk.LEFT)

        self.lbl_sliding_window = ttk.Label(custom_test_frame, text="")
        self.lbl_sliding_window.pack(anchor="w", padx=5, pady=(10, 0))
        lista_type_input = ['Yes', 'No']
        self.sliding_window = ttk.Combobox(custom_test_frame, values=lista_type_input, textvariable=self.type_input_var, state='readonly')
        self.sliding_window.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.sliding_window_hint = CreateToolTip(self.sliding_window, text='')

        self.button_ct = ttk.Button(custom_test_frame, text="", command=self.on_click_custom)
        self.button_ct.pack(fill=tk.X, padx=5, pady=10)
        self.prev_ct_hint = CreateToolTip(self.button_ct, text='')

        global_test_frame = ttk.LabelFrame(self.scrollable_frame_left, text="")
        global_test_frame.pack(fill=tk.X, padx=5, pady=5)
        self.global_test_frame_label = global_test_frame

        self.lbl_which_mls = ttk.Label(global_test_frame, text="")
        self.lbl_which_mls.pack(anchor="w", padx=5)
        self.wich_mls_hint = CreateToolTip(self.lbl_which_mls, text='')
        mls_check_frame = ttk.Frame(global_test_frame)
        mls_check_frame.pack(fill=tk.X, padx=5)
        self.ml_check_vars = {
            'NN': self.pre_nn_comb_var, 'DT': self.pre_dt_comb_var,
            'NNeig.': self.pre_nneig_comb_var, 'SV': self.pre_sv_comb_var,
            'GP': self.pre_gp_comb_var, 'BT': self.pre_bt_comb_var
        }
        self.ml_checkbuttons = {}
        for name, var in self.ml_check_vars.items():
            cb = ttk.Checkbutton(mls_check_frame, text=name, variable=var, onvalue=1, offvalue=0)
            cb.pack(side=tk.LEFT, padx=5)
            self.ml_checkbuttons[name] = cb

        self.lbl_indicator_gt = ttk.Label(global_test_frame, text="")
        self.lbl_indicator_gt.pack(anchor="w", padx=5, pady=(10, 0))
        self.indicator_global = ttk.Combobox(global_test_frame, values=lista_ind_meta_p, textvariable=self.ind_meta_comb_var, state='readonly')
        self.indicator_global.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.indicator_gt_hint = CreateToolTip(self.indicator_global, text='')

        self.lbl_num_tests_gt = ttk.Label(global_test_frame, text="")
        self.lbl_num_tests_gt.pack(anchor="w", padx=5)
        self.num_tests_global = ttk.Entry(global_test_frame, textvariable=self.num_teste_mtc_var, justify=tk.CENTER)
        self.num_tests_global.pack(fill=tk.X, padx=5, pady=(0, 10))
        self.num_tests_gt_hint = CreateToolTip(self.num_tests_global, text='')

        self.button_gt = ttk.Button(global_test_frame, text="", command=self.on_click_global)
        self.button_gt.pack(fill=tk.X, padx=5, pady=10)
        self.button_gt_hint = CreateToolTip(self.button_gt, text='')

        warning_frame = ttk.LabelFrame(self.scrollable_frame_left, text="")
        warning_frame.pack(fill=tk.X, padx=5, pady=5)
        self.warning_frame_label = warning_frame
        self.lbl_warning1 = ttk.Label(warning_frame, text="", wraplength=380, justify=tk.LEFT)
        self.lbl_warning1.pack(anchor="w", padx=5)
        self.lbl_warning2 = ttk.Label(warning_frame, text="", wraplength=380, justify=tk.LEFT)
        self.lbl_warning2.pack(anchor="w", padx=5)
        self.lbl_warning3 = ttk.Label(warning_frame, text="", wraplength=380, justify=tk.LEFT)
        self.lbl_warning3.pack(anchor="w", padx=5, pady=(0,5))

        # --- PAINEL DIREITO ---
        self.right_panel = ttk.Frame(main_container)
        self.right_panel.grid(row=0, column=1, sticky="nsew")

        self.update_texts()

    def go_to_previous_page(self):
        """Importa e navega para a página anterior."""
        from ..imputation_techniques_page import ImputationTechniquesPage
        self.controller.show_frame(ImputationTechniquesPage)

    def _clear_panel(self, panel):
        """Remove todos os widgets de um painel."""
        for widget in panel.winfo_children():
            widget.destroy()

    # --- Button Animation Logic (Adapted) ---
    def on_click_custom(self):
        self.button_ct.config(state=tk.DISABLED)
        self.loading_ct = True
        self.loading_step = 0
        self.animate_loading_ct()
        threading.Thread(target=self.run_process_custom, daemon=True).start()

    def animate_loading_ct(self):
        if self.loading_ct:
            dots = '.' * (self.loading_step % 4)
            i18n = self.controller.i18n
            loading_text = i18n.get('loading_text')
            self.button_ct.config(text=f"{loading_text}{dots}")
            self.loading_step += 1
            self.after(500, self.animate_loading_ct)

    def run_process_custom(self):
        try:
            results = self._generate_custom_test_logic()
            self.after(0, self._display_custom_results, results)
        except Exception as e:
            self.after(0, self.controller.show_translated_message(
                msg_type='error',
                title_key='customized_test_error_title',
                message_key='error_msg',
                error=str(e)
                ))
        finally:
            self.after(0, self.reset_button_custom)

    def reset_button_custom(self):
        self.loading_ct = False
        i18n = self.controller.i18n
        self.button_ct.config(text=i18n.get('generate_preview_btn'), state=tk.NORMAL)

    def on_click_global(self):
        self.button_gt.config(state=tk.DISABLED)
        self.loading_gt = True
        self.loading_step = 0
        self.animate_loading_gt()
        threading.Thread(target=self.run_process_global, daemon=True).start()

    def animate_loading_gt(self):
        if self.loading_gt:
            dots = '.' * (self.loading_step % 4)
            i18n = self.controller.i18n
            loading_text = i18n.get('loading_text')
            self.button_gt.config(text=f"{loading_text}{dots}")
            self.loading_step += 1
            self.after(500, self.animate_loading_gt)

    def run_process_global(self):
        try:
            results = self._generate_global_test_logic()
            self.after(0, self._display_global_results, results)
        except Exception as e:
             self.after(0, self.controller.show_translated_message(
                msg_type='error',
                title_key='global_test_error_title',
                message_key='error_msg',
                ))
        finally:
            self.after(0, self.reset_button_global)

    def reset_button_global(self):
        self.loading_gt = False
        i18n = self.controller.i18n
        self.button_gt.config(text=i18n.get('generate_preview_gt_btn'), state=tk.NORMAL)


    def _generate_custom_test_logic(self):
        """Coleta inputs e chama a lógica de meta-learning customizado."""
        base_model = self.ml_lv0_var.get()
        triangulation = self.ml_tr0_var.get()
        meta_model = self.ml_lv1_var.get()
        indicator = self.ind_meta_perso_var.get()
        num_tests = self.num_teste_mtp_var.get()
        pre_level_0 = self.pre_para_lv0_var.get()
        pre_level_1 = self.pre_para_lv1_var.get()
        input_window = self.type_input_var.get()

        if indicator == "Precipitation": focus = 1
        elif indicator == 'Maximum temperature': focus = 2
        else: focus = 3

        meta = MetaLearning()
        results_tuple = meta.customized_meta_learning(
            focus, base_model, triangulation, meta_model,
            pre_level_0, pre_level_1, num_tests, input_window
        )
        return results_tuple

    def _generate_global_test_logic(self):
        """Coleta inputs e chama a lógica de meta-learning global."""
        target_variable = self.ind_meta_comb_var.get()
        num_tests = self.num_teste_mtc_var.get()
        window_type = 'Yes'

        selected_models = [name for name, var in self.ml_check_vars.items() if var.get() == 1]
        if not selected_models:
            raise ValueError("Nenhum modelo de ML selecionado para o teste global.")

        meta = MetaLearning()
        all_models, model_ranking = meta.combine_meta_learning(
            target_variable, 0, 0, num_tests, window_type, selected_models
        )
        return all_models, model_ranking


    def _display_custom_results(self, results):
        """Exibe os resultados do teste customizado no painel direito."""
        self._clear_panel(self.right_panel)
        if results is None: return

        (meta_ea, meta_er, meta_percent_error, meta_r2, x_meta, y_meta, y_target,
         base_ea, base_er, base_percent, base_r2, tria_ea, tria_er) = results

        i18n = self.controller.i18n

        results_frame = ttk.LabelFrame(self.right_panel, text=i18n.get('results_preview_label'))
        results_frame.pack(side=tk.TOP, fill=tk.X, padx=10, pady=(10, 5))
        stats_frame = ttk.Frame(results_frame)
        stats_frame.pack(padx=5, pady=5, fill=tk.X)

        # Labels com as métricas
        errors = [
            (i18n.get('abs_error_label'), base_ea, tria_ea, meta_ea),
            (i18n.get('rel_error_label'), base_er, tria_er, meta_er),
            (i18n.get('perc_error_label'), base_percent, tria_ea * 100 if tria_ea is not None else None, meta_percent_error),
            (i18n.get('r2_label'), base_r2, None, meta_r2)
        ]

        for idx, (label, val_ml, val_tri, val_meta) in enumerate(errors):
            row_frame = ttk.Frame(stats_frame)
            row_frame.pack(fill=tk.X)
            ttk.Label(row_frame, text=f"{label}:", width=15, anchor="w").pack(side=tk.LEFT, padx=(0,5))
            if val_ml is not None:
                 ttk.Label(row_frame, text=f"{i18n.get('ml_label')}: {round(val_ml, 4)}", width=20).pack(side=tk.LEFT, padx=5)
            if val_tri is not None:
                 ttk.Label(row_frame, text=f"{i18n.get('triang_label')}: {round(val_tri, 4)}", width=20).pack(side=tk.LEFT, padx=5)
            if val_meta is not None:
                 ttk.Label(row_frame, text=f"{i18n.get('meta_label')}: {round(val_meta, 4)}", width=20).pack(side=tk.LEFT, padx=5)


        figure = Figure(figsize=(10, 6), dpi=100) 

        # Absolute Error
        ax1 = figure.add_subplot(2, 2, 1)
        bars1_x = [i18n.get('ml_label'), i18n.get('triang_label'), i18n.get('meta_label')]
        bars1_y = [base_ea, tria_ea, meta_ea]
        ax1.bar(bars1_x, bars1_y)
        ax1.set_ylabel(i18n.get('abs_error_header'))

        # Relative Error
        ax2 = figure.add_subplot(2, 2, 2)
        bars2_y = [base_er, tria_er, meta_er]
        ax2.bar(bars1_x, bars2_y) # Reusa x labels
        ax2.set_ylabel(i18n.get('rel_error_header'))

        # Percentage Error
        ax3 = figure.add_subplot(2, 2, 3)
        bars3_y = [base_percent, tria_ea * 100 if tria_ea is not None else 0, meta_percent_error]
        ax3.bar(bars1_x, bars3_y)
        ax3.set_ylabel(i18n.get('perc_error_header'))

        # R² Score
        ax4 = figure.add_subplot(2, 2, 4)
        bars4_x = [i18n.get('ml_label'), i18n.get('meta_label')]
        bars4_y = [base_r2, meta_r2]
        ax4.bar(bars4_x, bars4_y)
        ax4.set_ylabel(i18n.get('r2_label'))

        figure.tight_layout()

        canvas = FigureCanvasTkAgg(figure, master=self.right_panel)
        canvas.draw()
        toolbar = NavigationToolbar2Tk(canvas, self.right_panel)
        toolbar.update()
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True, padx=10, pady=(0, 10))
        toolbar.pack(side=tk.BOTTOM, fill=tk.X, padx=10, pady=(0, 10))

    def _display_global_results(self, results):
        """Exibe os resultados do teste global (tabelas e gráfico) no painel direito."""
        self._clear_panel(self.right_panel)
        if results is None: return

        all_models, model_ranking = results
        i18n = self.controller.i18n

        all_models_frame = ttk.LabelFrame(self.right_panel, text=i18n.get('generated_models_label'))
        all_models_frame.pack(fill=tk.BOTH, expand=True, padx=10, pady=(10, 5))

        all_data = [[m[0], m[1], m[2], m[3], m[5], m[6], m[7]] for m in all_models]
        headers = [
            i18n.get('model_header'), i18n.get('base_learning_header'),
            i18n.get('triangulation_header'), i18n.get('meta_learning_header'),
            i18n.get('abs_error_header'), i18n.get('rel_error_header'),
            i18n.get('perc_error_header')
        ]

        all_models_table = Sheet(all_models_frame, data=all_data, headers=headers)
        all_models_table.enable_bindings()
        all_models_table.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)

        # --- Ranking e Gráfico ---
        bottom_frame = ttk.Frame(self.right_panel)
        bottom_frame.pack(fill=tk.BOTH, expand=True, padx=10, pady=(5, 10))
        bottom_frame.grid_columnconfigure(0, weight=1) # Tabela Ranking
        bottom_frame.grid_columnconfigure(1, weight=3) # Gráfico Ranking
        bottom_frame.grid_rowconfigure(0, weight=1)

        # Frame para a Tabela de Ranking
        ranking_frame = ttk.LabelFrame(bottom_frame, text=i18n.get('model_ranking_label'))
        ranking_frame.grid(row=0, column=0, sticky="nsew", padx=(0, 5))

        ranking_data = [[m[0], round(float(str(m[1]).replace(',', '.')), 4)] for m in model_ranking]
        ranking_headers = [i18n.get('model_header'), i18n.get('error_header')]
        ranking_table = Sheet(ranking_frame, data=ranking_data, headers=ranking_headers, width=250)
        ranking_table.enable_bindings()
        ranking_table.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)

        graph_frame = ttk.LabelFrame(bottom_frame, text=i18n.get('model_ranking_label'))
        graph_frame.grid(row=0, column=1, sticky="nsew", padx=(5, 0))

        top_n = 15
        x_labels = [str(m[0]) for i, m in enumerate(model_ranking) if i < top_n]
        y_values = [round(float(str(m[1]).replace(',', '.')), 4) for i, m in enumerate(model_ranking) if i < top_n]

        figure = Figure(figsize=(8, 3), dpi=100)
        plot = figure.add_subplot(111)
        plot.bar(x_labels, y_values)
        plot.set_ylabel(i18n.get('perc_error_header'))
        plot.tick_params(axis='x', rotation=45, labelsize=8)
        plot.grid(True, axis='y')
        figure.tight_layout()

        canvas = FigureCanvasTkAgg(figure, master=graph_frame)
        canvas.draw()
        toolbar = NavigationToolbar2Tk(canvas, graph_frame)
        toolbar.update()
        canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True, padx=5, pady=(0, 5))
        toolbar.pack(side=tk.BOTTOM, fill=tk.X, padx=5, pady=(0, 5))

    def update_texts(self):
        """Atualiza os textos estáticos da tela."""
        i18n = self.controller.i18n
        self.back_button.config(text=i18n.get('back_btn'))
        self.page_title.config(text=i18n.get('meta_learning_page_title'))
        self.custom_test_frame_label.config(text=i18n.get('custom_test_label'))
        self.lbl_base_learning.config(text=i18n.get('base_learning_l0_label'))
        self.lbl_triangulation.config(text=i18n.get('triangulation_l0_label'))
        self.lbl_meta_learning.config(text=i18n.get('meta_learning_l1_label'))
        self.lbl_indicator_ct.config(text=i18n.get('climatic_indicator_label'))
        self.lbl_num_tests_ct.config(text=i18n.get('num_tests_label'))
        self.lbl_pretrained.config(text=i18n.get('use_pretrained_label'))
        self.cb_pre_lv0.config(text=i18n.get('level0_label'))
        self.cb_pre_lv1.config(text=i18n.get('level1_label'))
        self.lbl_sliding_window.config(text=i18n.get('use_sliding_window_label'))
        self.button_ct.config(text=i18n.get('generate_preview_btn'))
        self.global_test_frame_label.config(text=i18n.get('global_test_label'))
        self.lbl_which_mls.config(text=i18n.get('which_mls_label'))
        self.lbl_indicator_gt.config(text=i18n.get('climatic_indicator_label'))
        self.lbl_num_tests_gt.config(text=i18n.get('num_tests_label'))
        self.button_gt.config(text=i18n.get('generate_preview_gt_btn'))
        self.warning_frame_label.config(text=i18n.get('warning_label'))
        self.lbl_warning1.config(text=i18n.get('warning_text_1'))
        self.lbl_warning2.config(text=i18n.get('warning_text_2'))
        self.lbl_warning3.config(text=i18n.get('warning_text_3'))
        if not self.loading_ct:
             self.button_ct.config(text=i18n.get('generate_preview_btn'))
        if not self.loading_gt:
             self.button_gt.config(text=i18n.get('generate_preview_gt_btn'))

        self.ml0_hint.text = i18n.get('ml0_MeL_hint')
        self.tr0_hint.text = i18n.get('tr0_MeL_hint')
        self.ml1_hint.text = i18n.get('ml1_MeL_hint')
        self.indicator_ct_hint.text = i18n.get('ct_ind_MeL_hint')
        self.num_tests_hint.text = i18n.get('num_tests_MeL_hint')
        self.pretrained_hint.text = i18n.get('pretrained_MeL_hint')
        self.sliding_window_hint.text = i18n.get('sliding_window_MeL_hint')
        self.prev_ct_hint.text = i18n.get('prev_ct_MeL_hint')
        self.wich_mls_hint.text = i18n.get('wich_ml_MeL_hint')
        self.indicator_gt_hint.text = i18n.get('indicator_gt_MeL_hint')
        self.num_tests_gt_hint.text = i18n.get('num_tests_gt_MeL_hint')
        self.button_gt_hint.text = i18n.get('btn_gt_MeL_hint')



INDICATOR_MAP = {
    "Precipitation": 3,
    'Maximum temperature': 4,
    'Minimum temperature': 5
}

def get_indicator_code(indicator_name: str) -> int:
    return INDICATOR_MAP.get(indicator_name, -1)  # Retorna -1 se não encontrado


import tkinter as tk
from tkinter import ttk, StringVar, IntVar, BooleanVar

class SVMParameterFrame(ttk.Frame):
    """
    Frame com parâmetros para Support Vector Machine (Regression).
    """
    def __init__(self, master):
        super().__init__(master)

        # Configura as colunas da grade para expandirem igualmente
        self.grid_columnconfigure(0, weight=1, uniform="group1")
        self.grid_columnconfigure(1, weight=1, uniform="group1")

        self._create_widgets()

    def _create_widgets(self):
        # --- Linha 0: Kernel e Degree ---
        ttk.Label(self, text='Kernel:').grid(row=0, column=0, sticky="w", padx=5)
        self.kernel_v = StringVar(value='rbf')
        lista_ker = ['linear', 'poly', 'rbf', 'sigmoid']
        ttk.Combobox(self, values=lista_ker, textvariable=self.kernel_v, state='readonly').grid(row=1, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Degree (int, para poly):').grid(row=0, column=1, sticky="w", padx=5)
        self.degree_v = IntVar(value=3)
        ttk.Entry(self, textvariable=self.degree_v, justify=tk.CENTER).grid(row=1, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 2: Gamma e Coef0 ---
        ttk.Label(self, text='Gamma ("scale", "auto", float):').grid(row=2, column=0, sticky="w", padx=5)
        self.gamma_v = StringVar(value='scale')
        ttk.Entry(self, textvariable=self.gamma_v, justify=tk.CENTER).grid(row=3, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Coef0 (float, para poly/sigmoid):').grid(row=2, column=1, sticky="w", padx=5)
        self.coef0_v = StringVar(value='0.0')
        ttk.Entry(self, textvariable=self.coef0_v, justify=tk.CENTER).grid(row=3, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 4: Tol e C ---
        ttk.Label(self, text='Tol (float):').grid(row=4, column=0, sticky="w", padx=5)
        self.tol_v = StringVar(value='0.001')
        ttk.Entry(self, textvariable=self.tol_v, justify=tk.CENTER).grid(row=5, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='C (float, Regularization):').grid(row=4, column=1, sticky="w", padx=5)
        self.c_v = StringVar(value='1.0')
        ttk.Entry(self, textvariable=self.c_v, justify=tk.CENTER).grid(row=5, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 6: Epsilon e Shrinking ---
        ttk.Label(self, text='Epsilon (float):').grid(row=6, column=0, sticky="w", padx=5)
        self.epsilon_v = StringVar(value='0.1')
        ttk.Entry(self, textvariable=self.epsilon_v, justify=tk.CENTER).grid(row=7, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Shrinking (Bool):').grid(row=6, column=1, sticky="w", padx=5)
        self.shrinking_v = BooleanVar(value=True)
        ttk.Checkbutton(self, variable=self.shrinking_v, onvalue=True, offvalue=False).grid(row=7, column=1, sticky="w", padx=5, pady=(0, 10))

        # --- Linha 8: Cache Size e Verbose ---
        ttk.Label(self, text='Cache_size (MB, float):').grid(row=8, column=0, sticky="w", padx=5)
        self.cache_size_v = StringVar(value='200')
        ttk.Entry(self, textvariable=self.cache_size_v, justify=tk.CENTER).grid(row=9, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Verbose (Bool):').grid(row=8, column=1, sticky="w", padx=5)
        self.verbose_v = BooleanVar(value=False)
        ttk.Checkbutton(self, variable=self.verbose_v, onvalue=True, offvalue=False).grid(row=9, column=1, sticky="w", padx=5, pady=(0, 10))

        # --- Linha 10: Max Iter ---
        ttk.Label(self, text='Max_iter (int, -1 for no limit):').grid(row=10, column=0, sticky="w", padx=5)
        self.maxiter_v = IntVar(value=-1)
        ttk.Entry(self, textvariable=self.maxiter_v, justify=tk.CENTER).grid(row=11, column=0, sticky="ew", padx=5, pady=(0, 10))

class NNeighParameterFrame(ttk.Frame):
    """
    Frame com parâmetros para Nearest Neighbors.
    """
    def __init__(self, master):
        super().__init__(master)
        
        self.grid_columnconfigure(0, weight=1, uniform="group1")
        self.grid_columnconfigure(1, weight=1, uniform="group1")
        
        self._create_widgets()

    def _create_widgets(self):
        # --- Linha 0 ---
        ttk.Label(self, text='N_neighbors (int):').grid(row=0, column=0, sticky="w", padx=5)
        ttk.Label(self, text='Algorithm:').grid(row=0, column=1, sticky="w", padx=5)
        
        self.n_neighbors_v = IntVar(value=5)
        ttk.Entry(self, textvariable=self.n_neighbors_v, justify=tk.CENTER).grid(row=1, column=0, sticky="ew", padx=5, pady=(0, 10))

        self.algorithm_v = StringVar(value='auto')
        lista_alg = ['auto', 'ball_tree', 'kd_tree', 'brute']
        ttk.Combobox(self, values=lista_alg, textvariable=self.algorithm_v, state='readonly').grid(row=1, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 2 ---
        ttk.Label(self, text='Leaf_size (int):').grid(row=2, column=0, sticky="w", padx=5)
        ttk.Label(self, text='P (int):').grid(row=2, column=1, sticky="w", padx=5)

        self.leaf_size_v = IntVar(value=30)
        ttk.Entry(self, textvariable=self.leaf_size_v, justify=tk.CENTER).grid(row=3, column=0, sticky="ew", padx=5, pady=(0, 10))

        self.p_v = IntVar(value=2)
        ttk.Entry(self, textvariable=self.p_v, justify=tk.CENTER).grid(row=3, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 4 ---
        ttk.Label(self, text='N_jobs (int / "None"):').grid(row=4, column=0, sticky="w", padx=5)
        self.n_jobs_v = StringVar(value='-1')
        ttk.Entry(self, textvariable=self.n_jobs_v, justify=tk.CENTER).grid(row=5, column=0, sticky="ew", padx=5, pady=(0, 10))

class GPParameterFrame(ttk.Frame):
    """
    Frame com parâmetros para Gaussian Process.
    Agora com lógica para habilitar/desabilitar campos baseados no Kernel.
    """
    def __init__(self, master):
        super().__init__(master)
        
        self.grid_columnconfigure(0, weight=1, uniform="group1")
        self.grid_columnconfigure(1, weight=1, uniform="group1")
        
        self._create_widgets()
        self._update_kernel_inputs()

    def _create_widgets(self):
        ttk.Label(self, text='Alpha (Noise Level):').grid(row=0, column=0, sticky="w", padx=5)
        ttk.Label(self, text='N Restarts Optimizer:').grid(row=0, column=1, sticky="w", padx=5)

        self.alpha_gp = StringVar(value='1e-10') 
        ttk.Entry(self, textvariable=self.alpha_gp, justify=tk.CENTER).grid(row=1, column=0, sticky="ew", padx=5, pady=(0, 10))

        self.n_restarts_op = IntVar(value=1)
        ttk.Entry(self, textvariable=self.n_restarts_op, justify=tk.CENTER).grid(row=1, column=1, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Normalize Y:').grid(row=2, column=0, sticky="w", padx=5)
        ttk.Label(self, text='Kernel Type:').grid(row=2, column=1, sticky="w", padx=5)

        self.normalize_y_gp = BooleanVar(value=True) 
        ttk.Checkbutton(self, text="Sim/Não", variable=self.normalize_y_gp).grid(row=3, column=0, sticky="w", padx=5, pady=(0, 10))

        self.kernel_type = StringVar(value='RBF')
        self.combo_kernel = ttk.Combobox(self, textvariable=self.kernel_type, 
                                         values=["RBF", "Matern", "DotProduct", "RationalQuadratic"], 
                                         state='readonly')
        self.combo_kernel.grid(row=3, column=1, sticky="ew", padx=5, pady=(0, 10))
        
        self.combo_kernel.bind("<<ComboboxSelected>>", self._update_kernel_inputs)

        self.kernel_frame = ttk.LabelFrame(self, text="Parâmetros Específicos do Kernel")
        self.kernel_frame.grid(row=4, column=0, columnspan=2, sticky="ew", padx=5, pady=10)
        self.kernel_frame.grid_columnconfigure(0, weight=1)
        self.kernel_frame.grid_columnconfigure(1, weight=1)

        self.lbl_len = ttk.Label(self.kernel_frame, text='Length Scale:')
        self.lbl_len.grid(row=0, column=0, sticky="w", padx=5)
        self.length_scale = StringVar(value='1.0')
        self.ent_len = ttk.Entry(self.kernel_frame, textvariable=self.length_scale, justify=tk.CENTER)
        self.ent_len.grid(row=1, column=0, sticky="ew", padx=5, pady=(0, 5))
        
        self.lbl_nu = ttk.Label(self.kernel_frame, text='Nu (Matern):')
        self.lbl_nu.grid(row=0, column=1, sticky="w", padx=5)
        self.nu = StringVar(value='1.5')
        self.ent_nu = ttk.Combobox(self.kernel_frame, textvariable=self.nu, values=["0.5", "1.5", "2.5"], state="readonly")
        self.ent_nu.grid(row=1, column=1, sticky="ew", padx=5, pady=(0, 5))

        self.lbl_sigma = ttk.Label(self.kernel_frame, text='Sigma 0 (DotProduct):')
        self.lbl_sigma.grid(row=2, column=0, sticky="w", padx=5)
        self.sigma_0 = StringVar(value='1.0')
        self.ent_sigma = ttk.Entry(self.kernel_frame, textvariable=self.sigma_0, justify=tk.CENTER)
        self.ent_sigma.grid(row=3, column=0, sticky="ew", padx=5, pady=(0, 5))

        self.lbl_alpha_rq = ttk.Label(self.kernel_frame, text='Alpha (RationalQuadratic):')
        self.lbl_alpha_rq.grid(row=2, column=1, sticky="w", padx=5)
        self.alpha_rq = StringVar(value='1.0')
        self.ent_rq = ttk.Entry(self.kernel_frame, textvariable=self.alpha_rq, justify=tk.CENTER)
        self.ent_rq.grid(row=3, column=1, sticky="ew", padx=5, pady=(0, 5))

    def _update_kernel_inputs(self, event=None):
        """Habilita ou desabilita campos baseados na escolha do Kernel."""
        k_type = self.kernel_type.get()

        state_len = 'disabled'
        state_nu = 'disabled'
        state_sigma = 'disabled'
        state_rq = 'disabled'

        if k_type == 'RBF':
            state_len = 'normal'
        elif k_type == 'Matern':
            state_len = 'normal'
            state_nu = 'readonly'
        elif k_type == 'RationalQuadratic':
            state_len = 'normal'
            state_rq = 'normal'
        elif k_type == 'DotProduct':
            state_sigma = 'normal'

        self.ent_len.config(state=state_len)
        self.ent_nu.config(state=state_nu)
        self.ent_sigma.config(state=state_sigma)
        self.ent_rq.config(state=state_rq)

class BaggingTParameterFrame(ttk.Frame):
    """
    Frame com parâmetros para Bagging Trees.
    """
    def __init__(self, master):
        super().__init__(master)
        
        self.grid_columnconfigure(0, weight=1, uniform="group1")
        self.grid_columnconfigure(1, weight=1, uniform="group1")
        
        self._create_widgets()

    def _create_widgets(self):
        # --- Linha 0 ---
        ttk.Label(self, text='Criterion:').grid(row=0, column=0, sticky="w", padx=5)
        self.criterion_v = StringVar(value="absolute_error")
        lista_cri = ["squared_error", "friedman_mse", "absolute_error", "poisson"]
        ttk.Combobox(self, values=lista_cri, textvariable=self.criterion_v, state='readonly').grid(row=1, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Splitter:').grid(row=0, column=1, sticky="w", padx=5)
        self.splitter_v = StringVar(value="random")
        lista_spl = ["best", "random"]
        ttk.Combobox(self, values=lista_spl, textvariable=self.splitter_v, state='readonly').grid(row=1, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 2 ---
        ttk.Label(self, text="Max_depth (int):").grid(row=2, column=0, sticky="w", padx=5)
        self.maxd_v = StringVar(value="22")
        ttk.Entry(self, textvariable=self.maxd_v, justify=tk.CENTER).grid(row=3, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text="Min_samples_split (int):").grid(row=2, column=1, sticky="w", padx=5)
        self.minsam_s_v = IntVar(value=6)
        ttk.Entry(self, textvariable=self.minsam_s_v, justify=tk.CENTER).grid(row=3, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 4 ---
        ttk.Label(self, text="Min_samples_leaf (int):").grid(row=4, column=0, sticky="w", padx=5)
        self.minsam_l_v = IntVar(value=5)
        ttk.Entry(self, textvariable=self.minsam_l_v, justify=tk.CENTER).grid(row=5, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text="Min_weight_fraction_leaf (float):").grid(row=4, column=1, sticky="w", padx=5)
        self.minweifra_l_v = StringVar(value="0.0")
        ttk.Entry(self, textvariable=self.minweifra_l_v, justify=tk.CENTER).grid(row=5, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 6 ---
        ttk.Label(self, text="Max_features (int/float/'sqrt'/'log2'):").grid(row=6, column=0, sticky="w", padx=5)
        self.maxfeat_v = StringVar(value="sqrt")
        ttk.Entry(self, textvariable=self.maxfeat_v, justify=tk.CENTER).grid(row=7, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text="Max_leaf_nodes (int):").grid(row=6, column=1, sticky="w", padx=5)
        self.maxleaf_n = StringVar(value="10")
        ttk.Entry(self, textvariable=self.maxleaf_n, justify=tk.CENTER).grid(row=7, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 8 ---
        ttk.Label(self, text="Min_impurity_decrease (float):").grid(row=8, column=0, sticky="w", padx=5)
        self.minimp_dec = StringVar(value="0.0")
        ttk.Entry(self, textvariable=self.minimp_dec, justify=tk.CENTER).grid(row=9, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text="Ccp_alpha (float >= 0.0):").grid(row=8, column=1, sticky="w", padx=5)
        self.ccp_alp_v = StringVar(value="0.001")
        ttk.Entry(self, textvariable=self.ccp_alp_v, justify=tk.CENTER).grid(row=9, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 10 (Específico do Bagging) ---
        ttk.Label(self, text='N_Estimators (int):').grid(row=10, column=0, sticky="w", padx=5)
        self.n_estimators = IntVar(value=10)
        ttk.Entry(self, textvariable=self.n_estimators, justify=tk.CENTER).grid(row=11, column=0, sticky="ew", padx=5, pady=(0, 10))


class DTParameterFrame(ttk.Frame):
    """
    Frame com parâmetros para Decision Trees.
    """
    def __init__(self, master):
        super().__init__(master)
        
        self.grid_columnconfigure(0, weight=1, uniform="group1")
        self.grid_columnconfigure(1, weight=1, uniform="group1")
        
        self._create_widgets()

    def _create_widgets(self):
        # --- Linha 0 ---
        ttk.Label(self, text='Criterion:').grid(row=0, column=0, sticky="w", padx=5)
        self.criterion_v = StringVar(value="absolute_error")
        lista_cri = ["squared_error", "friedman_mse", "absolute_error", "poisson"]
        ttk.Combobox(self, values=lista_cri, textvariable=self.criterion_v, state='readonly').grid(row=1, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Splitter:').grid(row=0, column=1, sticky="w", padx=5)
        self.splitter_v = StringVar(value="random")
        lista_spl = ["best", "random"]
        ttk.Combobox(self, values=lista_spl, textvariable=self.splitter_v, state='readonly').grid(row=1, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 2 ---
        ttk.Label(self, text="Max_depth (int):").grid(row=2, column=0, sticky="w", padx=5)
        self.maxd_v = StringVar(value="22")
        ttk.Entry(self, textvariable=self.maxd_v, justify=tk.CENTER).grid(row=3, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text="Min_samples_split (int):").grid(row=2, column=1, sticky="w", padx=5)
        self.minsam_s_v = IntVar(value=6)
        ttk.Entry(self, textvariable=self.minsam_s_v, justify=tk.CENTER).grid(row=3, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 4 ---
        ttk.Label(self, text="Min_samples_leaf (int):").grid(row=4, column=0, sticky="w", padx=5)
        self.minsam_l_v = IntVar(value=5)
        ttk.Entry(self, textvariable=self.minsam_l_v, justify=tk.CENTER).grid(row=5, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text="Min_weight_fraction_leaf (float):").grid(row=4, column=1, sticky="w", padx=5)
        self.minweifra_l_v = StringVar(value="0.0")
        ttk.Entry(self, textvariable=self.minweifra_l_v, justify=tk.CENTER).grid(row=5, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 6 ---
        ttk.Label(self, text="Max_features (int/float/'sqrt'/'log2'):").grid(row=6, column=0, sticky="w", padx=5)
        self.maxfeat_v = StringVar(value="sqrt")
        ttk.Entry(self, textvariable=self.maxfeat_v, justify=tk.CENTER).grid(row=7, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text="Max_leaf_nodes (int):").grid(row=6, column=1, sticky="w", padx=5)
        self.maxleaf_n = StringVar(value="10")
        ttk.Entry(self, textvariable=self.maxleaf_n, justify=tk.CENTER).grid(row=7, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 8 ---
        ttk.Label(self, text="Min_impurity_decrease (float):").grid(row=8, column=0, sticky="w", padx=5)
        self.minimp_dec = StringVar(value="0.0")
        ttk.Entry(self, textvariable=self.minimp_dec, justify=tk.CENTER).grid(row=9, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text="Ccp_alpha (float >= 0.0):").grid(row=8, column=1, sticky="w", padx=5)
        self.ccp_alp_v = StringVar(value="0.001")
        ttk.Entry(self, textvariable=self.ccp_alp_v, justify=tk.CENTER).grid(row=9, column=1, sticky="ew", padx=5, pady=(0, 10))

class NNParameterFrame(ttk.Frame):
    """
    Frame com parâmetros para Neural Network.
    """
    def __init__(self, master):
        super().__init__(master)

        self.grid_columnconfigure(0, weight=1, uniform="group1")
        self.grid_columnconfigure(1, weight=1, uniform="group1")

        self._create_widgets()

    def _create_widgets(self):
        # --- Linha 0 ---
        ttk.Label(self, text='Activation:').grid(row=0, column=0, sticky="w", padx=5)
        self.activation_v = StringVar(value='relu')
        ttk.Combobox(self, values=['identity', 'logistic', 'tanh', 'relu'],
                     textvariable=self.activation_v, state='readonly').grid(row=1, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Solver:').grid(row=0, column=1, sticky="w", padx=5)
        self.solver_v = StringVar(value='adam')
        lista_sol = ['lbfgs', 'sgd', 'adam']
        ttk.Combobox(self, values=lista_sol, textvariable=self.solver_v, state='readonly').grid(row=1, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 2 ---
        ttk.Label(self, text='Alpha (float):').grid(row=2, column=0, sticky="w", padx=5)
        self.alpha_v = StringVar(value='0.0001')
        ttk.Entry(self, textvariable=self.alpha_v, justify=tk.CENTER).grid(row=3, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Batch_size (int / "auto"):').grid(row=2, column=1, sticky="w", padx=5)
        self.batch_size_v = StringVar(value='auto')
        ttk.Entry(self, textvariable=self.batch_size_v, justify=tk.CENTER).grid(row=3, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 4 ---
        ttk.Label(self, text="Learning_rate:").grid(row=4, column=0, sticky="w", padx=5)
        self.learning_rate_v = StringVar(value='constant')
        lista_learn = ['constant', 'invscaling', 'adaptive']
        ttk.Combobox(self, values=lista_learn, textvariable=self.learning_rate_v, state='readonly').grid(row=5, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Learning_rate_init (float):').grid(row=4, column=1, sticky="w", padx=5)
        self.learning_rate_init_v = StringVar(value='0.001')
        ttk.Entry(self, textvariable=self.learning_rate_init_v, justify=tk.CENTER).grid(row=5, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 6 ---
        ttk.Label(self, text='Power_t (float):').grid(row=6, column=0, sticky="w", padx=5)
        self.power_t_v = StringVar(value='0.5')
        ttk.Entry(self, textvariable=self.power_t_v, justify=tk.CENTER).grid(row=7, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Max_iter (int):').grid(row=6, column=1, sticky="w", padx=5)
        self.max_iter_v = StringVar(value='200')
        ttk.Entry(self, textvariable=self.max_iter_v, justify=tk.CENTER).grid(row=7, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 8 ---
        ttk.Label(self, text='Shuffle (Bool 1/0):').grid(row=8, column=0, sticky="w", padx=5)
        self.shuffle_v = BooleanVar(value=True)
        # Usando Checkbutton para BooleanVar
        ttk.Checkbutton(self, variable=self.shuffle_v, onvalue=True, offvalue=False).grid(row=9, column=0, sticky="w", padx=5, pady=(0, 10))

        ttk.Label(self, text='Tol (float):').grid(row=8, column=1, sticky="w", padx=5)
        self.tol_v = StringVar(value='0.0001')
        ttk.Entry(self, textvariable=self.tol_v, justify=tk.CENTER).grid(row=9, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 10 ---
        ttk.Label(self, text='Verbose (Bool 1/0):').grid(row=10, column=0, sticky="w", padx=5)
        self.verbose_v = BooleanVar(value=False)
        ttk.Checkbutton(self, variable=self.verbose_v, onvalue=True, offvalue=False).grid(row=11, column=0, sticky="w", padx=5, pady=(0, 10))

        ttk.Label(self, text='Warm_start (Bool 1/0):').grid(row=10, column=1, sticky="w", padx=5)
        self.warm_start_v = BooleanVar(value=False)
        ttk.Checkbutton(self, variable=self.warm_start_v, onvalue=True, offvalue=False).grid(row=11, column=1, sticky="w", padx=5, pady=(0, 10))

        # --- Linha 12 ---
        ttk.Label(self, text='Momentum (float):').grid(row=12, column=0, sticky="w", padx=5)
        self.momentum_v = StringVar(value='0.9')
        ttk.Entry(self, textvariable=self.momentum_v, justify=tk.CENTER).grid(row=13, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Nesterovs_momentum (Bool 1/0):').grid(row=12, column=1, sticky="w", padx=5)
        self.nesterovs_momentum_v = BooleanVar(value=True)
        ttk.Checkbutton(self, variable=self.nesterovs_momentum_v, onvalue=True, offvalue=False).grid(row=13, column=1, sticky="w", padx=5, pady=(0, 10))

        # --- Linha 14 ---
        ttk.Label(self, text='Early_stopping (Bool 1/0):').grid(row=14, column=0, sticky="w", padx=5)
        self.early_stopping_v = BooleanVar(value=False)
        ttk.Checkbutton(self, variable=self.early_stopping_v, onvalue=True, offvalue=False).grid(row=15, column=0, sticky="w", padx=5, pady=(0, 10))

        ttk.Label(self, text='Validation_fraction (float):').grid(row=14, column=1, sticky="w", padx=5)
        self.validation_fraction_v = StringVar(value='0.1')
        ttk.Entry(self, textvariable=self.validation_fraction_v, justify=tk.CENTER).grid(row=15, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 16 ---
        ttk.Label(self, text='Beta_1 (float):').grid(row=16, column=0, sticky="w", padx=5)
        self.beta_1_v = StringVar(value='0.9')
        ttk.Entry(self, textvariable=self.beta_1_v, justify=tk.CENTER).grid(row=17, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='Beta_2 (float):').grid(row=16, column=1, sticky="w", padx=5)
        self.beta_2_v = StringVar(value='0.999')
        ttk.Entry(self, textvariable=self.beta_2_v, justify=tk.CENTER).grid(row=17, column=1, sticky="ew", padx=5, pady=(0, 10))

        # --- Linha 18 ---
        ttk.Label(self, text='N_iter_no_change (int):').grid(row=18, column=0, sticky="w", padx=5)
        self.n_iter_no_change_v = StringVar(value='10')
        ttk.Entry(self, textvariable=self.n_iter_no_change_v, justify=tk.CENTER).grid(row=19, column=0, sticky="ew", padx=5, pady=(0, 10))

        ttk.Label(self, text='max_fun (int):').grid(row=18, column=1, sticky="w", padx=5)
        self.max_fun_v = StringVar(value='15000')
        ttk.Entry(self, textvariable=self.max_fun_v, justify=tk.CENTER).grid(row=19, column=1, sticky="ew", padx=5, pady=(0, 10))


from ..data_processing.data_processing import DataProcessing

def choose_data(focus):
    print("Entrou utils choose_data")
    data_process = DataProcessing()
    if focus == 1: # Precipitation
        index = 6
        a = 3
        data = data_process.normalize_data(data_process.load_data_file('Common data'))
    elif focus == 2: # Maximum temperature
        index = 7
        a = 4
        data = data_process.load_data_file('Common data')
    elif focus == 3: #Maximum temperature
        index = 8
        a = 5
        data = data_process.load_data_file('Common data')
    
    return index, a, data


import pandas as pd
import os
import numpy as np

          
def process_csv(cities: list, input_dir:str, output_dir:str):
    os.makedirs(output_dir, exist_ok=True)
    columns = ["data", "prec", "tmax", "tmean", "tmin"]

    for city in cities:
        input_file = os.path.join(input_dir, city)
        output_file = os.path.join(output_dir, city)
        try:
            df = pd.read_csv(input_file, skiprows=11, header=None, names=columns, sep=";", decimal=",", index_col=False)
            
            # Substituir valores ausentes nas colunas climáticas por -99
            df[["prec", "tmax", "tmin", "tmean"]] = df[["prec", "tmax", "tmin", "tmean"]].fillna(-99)
            # Remover linhas com qualquer -99
            #df = df[(df["prec"] != -99) & (df["tmax"] != -99) & (df["tmin"] != -99 & (df["tmean"] != -99))]

            df["data"] = pd.to_datetime(df["data"], format="%Y-%m-%d", errors="coerce")

            df["year"] = df["data"].dt.year
            df["month"] = df["data"].dt.month
            df["day"] = df["data"].dt.day

            df.drop(columns=["data"], inplace=True)

            sorted_columns = ["year", "month", "day", "prec", "tmax", "tmin", "tmean"]
            df = df[sorted_columns]
            
            df.to_csv(output_file, index=False, sep=",", header=False)
            

        except FileNotFoundError:
            print(f"Erro: Arquivo {input_file} não encontrado.")
        except Exception as e:
            print(f"Erro ao ler {input_file}: {e}")

